@@ -27,14 +27,11 @@
 
 #include <glib/gi18n.h>
 
-#include <gio/gunixinputstream.h>
-
 #include "libglnx/libglnx.h"
 
 #include "flatpak-builtins.h"
 #include "flatpak-builtins-utils.h"
 #include "flatpak-utils.h"
-#include "flatpak-chain-input-stream.h"
 
 static gboolean opt_no_gpg_verify;
 static gboolean opt_do_gpg_verify;
@@ -44,12 +41,14 @@ static gboolean opt_do_deps;
 static gboolean opt_no_deps;
 static gboolean opt_if_not_exists;
 static gboolean opt_enable;
+static gboolean opt_oci;
 static gboolean opt_update_metadata;
 static gboolean opt_disable;
 static int opt_prio = -1;
 static char *opt_title;
 static char *opt_default_branch;
 static char *opt_url;
+static char *opt_collection_id = NULL;
 static gboolean opt_from;
 static char **opt_gpg_import;
 
@@ -77,98 +76,41 @@ static GOptionEntry common_options[] = {
   { "prio", 0, 0, G_OPTION_ARG_INT, &opt_prio, N_("Set priority (default 1, higher is more prioritized)"), N_("PRIORITY") },
   { "title", 0, 0, G_OPTION_ARG_STRING, &opt_title, N_("A nice name to use for this remote"), N_("TITLE") },
   { "default-branch", 0, 0, G_OPTION_ARG_STRING, &opt_default_branch, N_("Default branch to use for this remote"), N_("BRANCH") },
+#ifdef FLATPAK_ENABLE_P2P
+  { "collection-id", 0, 0, G_OPTION_ARG_STRING, &opt_collection_id, N_("Collection ID"), N_("COLLECTION-ID") },
+#endif  /* FLATPAK_ENABLE_P2P */
   { "gpg-import", 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &opt_gpg_import, N_("Import GPG key from FILE (- for stdin)"), N_("FILE") },
   { "disable", 0, 0, G_OPTION_ARG_NONE, &opt_disable, N_("Disable the remote"), NULL },
+  { "oci", 0, 0, G_OPTION_ARG_NONE, &opt_oci, N_("Add OCI registry"), NULL },
   { NULL }
 };
 
-static gboolean
-open_source_stream (GInputStream **out_source_stream,
-                    GCancellable  *cancellable,
-                    GError       **error)
-{
-  g_autoptr(GInputStream) source_stream = NULL;
-  guint n_keyrings = 0;
-  g_autoptr(GPtrArray) streams = NULL;
-
-  if (opt_gpg_import != NULL)
-    n_keyrings = g_strv_length (opt_gpg_import);
-
-  guint ii;
-
-  streams = g_ptr_array_new_with_free_func (g_object_unref);
-
-  for (ii = 0; ii < n_keyrings; ii++)
-    {
-      GInputStream *input_stream = NULL;
-
-      if (strcmp (opt_gpg_import[ii], "-") == 0)
-        {
-          input_stream = g_unix_input_stream_new (STDIN_FILENO, FALSE);
-        }
-      else
-        {
-          g_autoptr(GFile) file = g_file_new_for_commandline_arg (opt_gpg_import[ii]);
-          input_stream = G_INPUT_STREAM (g_file_read (file, cancellable, error));
-
-          if (input_stream == NULL)
-            {
-              g_prefix_error (error, "The file %s specified for --gpg-import was not found: ", opt_gpg_import[ii]);
-              return FALSE;
-            }
-        }
-
-      /* Takes ownership. */
-      g_ptr_array_add (streams, input_stream);
-    }
-
-  /* Chain together all the --keyring options as one long stream. */
-  source_stream = (GInputStream *) flatpak_chain_input_stream_new (streams);
-
-  *out_source_stream = g_steal_pointer (&source_stream);
-
-  return TRUE;
-}
-
-static GBytes *
-load_keys (GCancellable *cancellable,
-           GError      **error)
-{
-  g_autoptr(GInputStream) input_stream = NULL;
-  g_autoptr(GOutputStream) output_stream = NULL;
-  gssize n_bytes_written;
-
-  if (!open_source_stream (&input_stream, cancellable, error))
-    return FALSE;
-
-  output_stream = g_memory_output_stream_new_resizable ();
-
-  n_bytes_written = g_output_stream_splice (output_stream, input_stream,
-                                            G_OUTPUT_STREAM_SPLICE_CLOSE_SOURCE |
-                                            G_OUTPUT_STREAM_SPLICE_CLOSE_TARGET,
-                                            NULL, error);
-  if (n_bytes_written < 0)
-    return NULL;
-
-  return g_memory_output_stream_steal_as_bytes (G_MEMORY_OUTPUT_STREAM (output_stream));
-}
 
 static GKeyFile *
-get_config_from_opts (FlatpakDir *dir, const char *remote_name)
+get_config_from_opts (FlatpakDir *dir, const char *remote_name, gboolean *changed)
 {
-  GKeyFile *config = ostree_repo_copy_config (flatpak_dir_get_repo (dir));
+  OstreeRepo *repo;
+  GKeyFile *config;
   g_autofree char *group = g_strdup_printf ("remote \"%s\"", remote_name);
 
+  repo = flatpak_dir_get_repo (dir);
+  if (repo == NULL)
+    config = g_key_file_new ();
+  else
+    config = ostree_repo_copy_config (repo);
+
   if (opt_no_gpg_verify)
     {
       g_key_file_set_boolean (config, group, "gpg-verify", FALSE);
       g_key_file_set_boolean (config, group, "gpg-verify-summary", FALSE);
+      *changed = TRUE;
     }
 
   if (opt_do_gpg_verify)
     {
       g_key_file_set_boolean (config, group, "gpg-verify", TRUE);
       g_key_file_set_boolean (config, group, "gpg-verify-summary", TRUE);
+      *changed = TRUE;
     }
 
   if (opt_url)
@@ -177,35 +119,76 @@ get_config_from_opts (FlatpakDir *dir, const char *remote_name)
         g_key_file_set_string (config, group, "metalink", opt_url + strlen ("metalink="));
       else
         g_key_file_set_string (config, group, "url", opt_url);
+      *changed = TRUE;
+    }
+
+  if (opt_collection_id)
+    {
+      g_key_file_set_string (config, group, "collection-id", opt_collection_id);
+      g_key_file_set_boolean (config, group, "gpg-verify-summary", FALSE);
+      *changed = TRUE;
     }
 
   if (opt_title)
-    g_key_file_set_string (config, group, "xa.title", opt_title);
+    {
+      g_key_file_set_string (config, group, "xa.title", opt_title);
+      g_key_file_set_boolean (config, group, "xa.title-is-set", TRUE);
+      *changed = TRUE;
+    }
 
   if (opt_default_branch)
-    g_key_file_set_string (config, group, "xa.default-branch", opt_default_branch);
+    {
+      g_key_file_set_string (config, group, "xa.default-branch", opt_default_branch);
+      g_key_file_set_boolean (config, group, "xa.default-branch-is-set", TRUE);
+      *changed = TRUE;
+    }
 
   if (opt_no_enumerate)
-    g_key_file_set_boolean (config, group, "xa.noenumerate", TRUE);
+    {
+      g_key_file_set_boolean (config, group, "xa.noenumerate", TRUE);
+      *changed = TRUE;
+    }
 
   if (opt_do_enumerate)
-    g_key_file_set_boolean (config, group, "xa.noenumerate", FALSE);
+    {
+      g_key_file_set_boolean (config, group, "xa.noenumerate", FALSE);
+      *changed = TRUE;
+    }
 
   if (opt_no_deps)
-    g_key_file_set_boolean (config, group, "xa.nodeps", TRUE);
+    {
+      g_key_file_set_boolean (config, group, "xa.nodeps", TRUE);
+      *changed = TRUE;
+    }
 
   if (opt_do_deps)
-    g_key_file_set_boolean (config, group, "xa.nodeps", FALSE);
+    {
+      g_key_file_set_boolean (config, group, "xa.nodeps", FALSE);
+      *changed = TRUE;
+    }
 
   if (opt_disable)
-    g_key_file_set_boolean (config, group, "xa.disable", TRUE);
+    {
+      g_key_file_set_boolean (config, group, "xa.disable", TRUE);
+      *changed = TRUE;
+    }
   else if (opt_enable)
-    g_key_file_set_boolean (config, group, "xa.disable", FALSE);
+    {
+      g_key_file_set_boolean (config, group, "xa.disable", FALSE);
+      *changed = TRUE;
+    }
+
+  if (opt_oci)
+    {
+      g_key_file_set_boolean (config, group, "xa.oci", TRUE);
+      *changed = TRUE;
+    }
 
   if (opt_prio != -1)
     {
       g_autofree char *prio_as_string = g_strdup_printf ("%d", opt_prio);
       g_key_file_set_string (config, group, "xa.prio", prio_as_string);
+      *changed = TRUE;
     }
 
   return config;
@@ -232,14 +215,14 @@ load_options (const char *filename,
 
       if (bytes == NULL)
         {
-          g_printerr ("Can't load uri %s: %s\n", filename, error->message);
+          g_printerr (_("Can't load uri %s: %s\n"), filename, error->message);
           exit (1);
         }
 
       options_data = g_bytes_get_data (bytes, &options_size);
       if (!g_key_file_load_from_data (keyfile, options_data, options_size, 0, &error))
         {
-          g_printerr ("Can't load uri %s: %s\n", filename, error->message);
+          g_printerr (_("Can't load uri %s: %s\n"), filename, error->message);
           exit (1);
         }
     }
@@ -247,7 +230,7 @@ load_options (const char *filename,
     {
       if (!g_key_file_load_from_file (keyfile, filename, 0, &error))
         {
-          g_printerr ("Can't load file %s: %s\n", filename, error->message);
+          g_printerr (_("Can't load file %s: %s\n"), filename, error->message);
           exit (1);
         }
     }
@@ -255,7 +238,7 @@ load_options (const char *filename,
 
   if (!g_key_file_has_group (keyfile, FLATPAK_REPO_GROUP))
     {
-      g_printerr ("Invalid file format");
+      g_printerr (_("Invalid file format"));
       exit (1);
     }
 
@@ -263,7 +246,7 @@ load_options (const char *filename,
                                    FLATPAK_REPO_VERSION_KEY, NULL);
   if (version != NULL && strcmp (version, "1") != 0)
     {
-      g_printerr ("Invalid version %s, only 1 supported", version);
+      g_printerr (_("Invalid version %s, only 1 supported"), version);
       exit (1);
     }
 
@@ -272,6 +255,12 @@ load_options (const char *filename,
   if (str != NULL)
     opt_url = str;
 
+#ifdef FLATPAK_ENABLE_P2P
+  str = g_key_file_get_string (keyfile, FLATPAK_REPO_GROUP, FLATPAK_REPO_COLLECTION_ID_KEY, NULL);
+  if (str != NULL)
+    opt_collection_id = str;
+#endif  /* FLATPAK_ENABLE_P2P */
+
   str = g_key_file_get_locale_string (keyfile, FLATPAK_REPO_GROUP,
                                       FLATPAK_REPO_TITLE_KEY, NULL, NULL);
   if (str != NULL)
@@ -301,7 +290,7 @@ load_options (const char *filename,
       decoded = g_base64_decode (str, &decoded_len);
       if (decoded_len < 10) /* Check some minimal size so we don't get crap */
         {
-          g_printerr ("Invalid gpg key");
+          g_printerr (_("Invalid gpg key"));
           exit (1);
         }
 
@@ -311,52 +300,13 @@ load_options (const char *filename,
     }
 }
 
-static gboolean
-update_remote_with_extra_metadata (FlatpakDir* dir,
-                                   const char *remote,
-                                   GBytes *gpg_data,
-                                   GCancellable *cancellable,
-                                   GError **error)
-{
-  g_autofree char *title = NULL;
-  g_autofree char *default_branch = NULL;
-  g_autoptr(GKeyFile) config = NULL;
-
-  if (opt_title == NULL)
-    {
-      title = flatpak_dir_fetch_remote_title (dir,
-                                              remote,
-                                              NULL,
-                                              NULL);
-      if (title)
-        opt_title = title;
-    }
-
-    if (opt_default_branch == NULL)
-    {
-      default_branch = flatpak_dir_fetch_remote_default_branch (dir,
-                                                                remote,
-                                                                NULL,
-                                                                NULL);
-      if (default_branch)
-        opt_default_branch = default_branch;
-    }
-
-    if (title != NULL || default_branch != NULL)
-      {
-        config = get_config_from_opts (dir, remote);
-        return flatpak_dir_modify_remote (dir, remote, config, gpg_data, cancellable, error);
-      }
-
-    return TRUE;
-}
-
 gboolean
 flatpak_builtin_add_remote (int argc, char **argv,
                             GCancellable *cancellable, GError **error)
 {
   g_autoptr(GOptionContext) context = NULL;
-  g_autoptr(FlatpakDir) dir = NULL;
+  g_autoptr(GPtrArray) dirs = NULL;
+  FlatpakDir *dir;
   g_autoptr(GFile) file = NULL;
   g_auto(GStrv) remotes = NULL;
   g_autofree char *remote_url = NULL;
@@ -364,15 +314,21 @@ flatpak_builtin_add_remote (int argc, char **argv,
   const char *location = NULL;
   g_autoptr(GKeyFile) config = NULL;
   g_autoptr(GBytes) gpg_data = NULL;
+  gboolean changed = FALSE;
+  g_autoptr(GError) local_error = NULL;
 
   context = g_option_context_new (_("NAME LOCATION - Add a remote repository"));
   g_option_context_set_translation_domain (context, GETTEXT_PACKAGE);
 
   g_option_context_add_main_entries (context, common_options, NULL);
 
-  if (!flatpak_option_context_parse (context, add_options, &argc, &argv, 0, &dir, cancellable, error))
+  if (!flatpak_option_context_parse (context, add_options, &argc, &argv,
+                                     FLATPAK_BUILTIN_FLAG_ONE_DIR | FLATPAK_BUILTIN_FLAG_OPTIONAL_REPO,
+                                     &dirs, cancellable, error))
     return FALSE;
 
+  dir = g_ptr_array_index (dirs, 0);
+
   if (argc < 2)
     return usage_error (context, _("NAME must be specified"), error);
 
@@ -382,6 +338,16 @@ flatpak_builtin_add_remote (int argc, char **argv,
   if (argc > 3)
     return usage_error (context, _("Too many arguments"), error);
 
+#ifdef FLATPAK_ENABLE_P2P
+  if (opt_collection_id != NULL &&
+      !ostree_validate_collection_id (opt_collection_id, &local_error))
+    return flatpak_fail (error, _("‘%s’ is not a valid collection ID: %s"), opt_collection_id, local_error->message);
+
+  if (opt_collection_id != NULL &&
+      (opt_no_gpg_verify || opt_gpg_import == NULL || opt_gpg_import[0] == NULL))
+    return flatpak_fail (error, _("GPG verification is required if collections are enabled"));
+#endif  /* FLATPAK_ENABLE_P2P */
+
   remote_name = argv[1];
   location = argv[2];
 
@@ -414,15 +380,18 @@ flatpak_builtin_add_remote (int argc, char **argv,
       opt_url = remote_url;
     }
 
-  /* Default to gpg verify */
-  if (!opt_no_gpg_verify)
+  if (opt_oci && g_getenv ("FLATPAK_ENABLE_EXPERIMENTAL_OCI") == NULL)
+    return flatpak_fail (error, "flatpak remote-add --oci is currently unsupported and experimental, enable it by setting the FLATPAK_ENABLE_EXPERIMENTAL_OCI env var");
+
+  /* Default to gpg verify, except for --oci */
+  if (!opt_no_gpg_verify && !opt_oci)
     opt_do_gpg_verify = TRUE;
 
-  config = get_config_from_opts (dir, remote_name);
+  config = get_config_from_opts (dir, remote_name, &changed);
 
   if (opt_gpg_import != NULL)
     {
-      gpg_data = load_keys (cancellable, error);
+      gpg_data = flatpak_load_gpg_keys (opt_gpg_import, cancellable, error);
       if (gpg_data == NULL)
         return FALSE;
     }
@@ -435,19 +404,35 @@ flatpak_builtin_add_remote (int argc, char **argv,
     return FALSE;
 
   /* We can't retrieve the extra metadata until the remote has been added locally, since
-     ostree_repo_remote_fetch_summary() works with the repository's name, not its URL. */
-  return update_remote_with_extra_metadata (dir, remote_name, gpg_data, cancellable, error);
+     ostree_repo_remote_fetch_summary() works with the repository's name, not its URL.
+     Don't propagate IO failed errors here because we might just be offline - the
+     remote should already be usable. */
+  if (!flatpak_dir_update_remote_configuration (dir, remote_name, cancellable, &local_error))
+    {
+      if (local_error->domain == G_RESOLVER_ERROR ||
+          g_error_matches (local_error, G_IO_ERROR, G_IO_ERROR_FAILED))
+        {
+          g_printerr (_("Warning: Could not update extra metadata for '%s': %s\n"), remote_name, local_error->message);
+        }
+      else
+        {
+          g_propagate_error (error, g_steal_pointer (&local_error));
+          return FALSE;
+        }
+    }
+
+  return TRUE;
 }
 
 gboolean
 flatpak_complete_add_remote (FlatpakCompletion *completion)
 {
   g_autoptr(GOptionContext) context = NULL;
-  g_autoptr(FlatpakDir) dir = NULL;
 
   context = g_option_context_new ("");
   g_option_context_add_main_entries (context, common_options, NULL);
-  if (!flatpak_option_context_parse (context, add_options, &completion->argc, &completion->argv, 0, &dir, NULL, NULL))
+  if (!flatpak_option_context_parse (context, add_options, &completion->argc, &completion->argv,
+                                     FLATPAK_BUILTIN_FLAG_ONE_DIR, NULL, NULL, NULL))
     return FALSE;
 
   switch (completion->argc)
@@ -469,17 +454,20 @@ gboolean
 flatpak_builtin_modify_remote (int argc, char **argv, GCancellable *cancellable, GError **error)
 {
   g_autoptr(GOptionContext) context = NULL;
-  g_autoptr(FlatpakDir) dir = NULL;
+  g_autoptr(GPtrArray) dirs = NULL;
+  g_autoptr(FlatpakDir) preferred_dir = NULL;
   g_autoptr(GKeyFile) config = NULL;
   g_autoptr(GBytes) gpg_data = NULL;
   const char *remote_name;
+  gboolean changed = FALSE;
 
   context = g_option_context_new (_("NAME - Modify a remote repository"));
   g_option_context_set_translation_domain (context, GETTEXT_PACKAGE);
 
   g_option_context_add_main_entries (context, common_options, NULL);
 
-  if (!flatpak_option_context_parse (context, modify_options, &argc, &argv, 0, &dir, cancellable, error))
+  if (!flatpak_option_context_parse (context, modify_options, &argc, &argv,
+                                     FLATPAK_BUILTIN_FLAG_STANDARD_DIRS, &dirs, cancellable, error))
     return FALSE;
 
   if (argc < 2)
@@ -487,47 +475,52 @@ flatpak_builtin_modify_remote (int argc, char **argv, GCancellable *cancellable,
 
   remote_name = argv[1];
 
-  if (!ostree_repo_remote_get_url (flatpak_dir_get_repo (dir), remote_name, NULL, NULL))
-    return flatpak_fail (error, _("No remote %s"), remote_name);
+  if (!flatpak_resolve_duplicate_remotes (dirs, remote_name, &preferred_dir, cancellable, error))
+    return FALSE;
 
   if (opt_update_metadata)
     {
       g_autoptr(GError) local_error = NULL;
 
       g_print (_("Updating extra metadata from remote summary for %s\n"), remote_name);
-      if (!flatpak_dir_update_remote_configuration (dir, remote_name, cancellable, &local_error))
+      if (!flatpak_dir_update_remote_configuration (preferred_dir, remote_name, cancellable, &local_error))
         {
           g_printerr (_("Error updating extra metadata for '%s': %s\n"), remote_name, local_error->message);
           return flatpak_fail (error, _("Could not update extra metadata for %s"), remote_name);
         }
 
       /* Reload changed configuration */
-      if (!flatpak_dir_recreate_repo (dir, cancellable, error))
+      if (!flatpak_dir_recreate_repo (preferred_dir, cancellable, error))
         return FALSE;
     }
 
-  config = get_config_from_opts (dir, remote_name);
+  config = get_config_from_opts (preferred_dir, remote_name, &changed);
 
   if (opt_gpg_import != NULL)
     {
-      gpg_data = load_keys (cancellable, error);
+      gpg_data = flatpak_load_gpg_keys (opt_gpg_import, cancellable, error);
       if (gpg_data == NULL)
         return FALSE;
+      changed = TRUE;
     }
 
-  return flatpak_dir_modify_remote (dir, remote_name, config, gpg_data, cancellable, error);
+  if (!changed)
+    return TRUE;
+
+  return flatpak_dir_modify_remote (preferred_dir, remote_name, config, gpg_data, cancellable, error);
 }
 
 gboolean
 flatpak_complete_modify_remote (FlatpakCompletion *completion)
 {
   g_autoptr(GOptionContext) context = NULL;
-  g_autoptr(FlatpakDir) dir = NULL;
+  g_autoptr(GPtrArray) dirs = NULL;
   int i;
 
   context = g_option_context_new ("");
   g_option_context_add_main_entries (context, common_options, NULL);
-  if (!flatpak_option_context_parse (context, modify_options, &completion->argc, &completion->argv, 0, &dir, NULL, NULL))
+  if (!flatpak_option_context_parse (context, modify_options, &completion->argc, &completion->argv,
+                                     FLATPAK_BUILTIN_FLAG_STANDARD_DIRS, &dirs, NULL, NULL))
     return FALSE;
 
   switch (completion->argc)
@@ -539,13 +532,16 @@ flatpak_complete_modify_remote (FlatpakCompletion *completion)
       flatpak_complete_options (completion, modify_options);
       flatpak_complete_options (completion, user_entries);
 
-      {
-        g_auto(GStrv) remotes = flatpak_dir_list_remotes (dir, NULL, NULL);
-        if (remotes == NULL)
-          return FALSE;
-        for (i = 0; remotes[i] != NULL; i++)
-          flatpak_complete_word (completion, "%s ", remotes[i]);
-      }
+      for (i = 0; i < dirs->len; i++)
+        {
+          FlatpakDir *dir = g_ptr_array_index (dirs, i);
+          int j;
+          g_auto(GStrv) remotes = flatpak_dir_list_remotes (dir, NULL, NULL);
+          if (remotes == NULL)
+            return FALSE;
+          for (j = 0; remotes[j] != NULL; j++)
+            flatpak_complete_word (completion, "%s ", remotes[j]);
+        }
 
       break;
     }
