@@ -44,6 +44,10 @@ static char **opt_include;
 static char *opt_gpg_homedir;
 static char *opt_files;
 static char *opt_metadata;
+static char *opt_timestamp = NULL;
+#ifdef FLATPAK_ENABLE_P2P
+static char *opt_collection_id = NULL;
+#endif  /* FLATPAK_ENABLE_P2P */
 
 static GOptionEntry options[] = {
   { "subject", 's', 0, G_OPTION_ARG_STRING, &opt_subject, N_("One line subject"), N_("SUBJECT") },
@@ -58,6 +62,10 @@ static GOptionEntry options[] = {
   { "exclude", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_exclude, N_("Files to exclude"), N_("PATTERN") },
   { "include", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_include, N_("Excluded files to include"), N_("PATTERN") },
   { "gpg-homedir", 0, 0, G_OPTION_ARG_STRING, &opt_gpg_homedir, N_("GPG Homedir to use when looking for keyrings"), N_("HOMEDIR") },
+  { "timestamp", 0, 0, G_OPTION_ARG_STRING, &opt_timestamp, N_("Override the timestamp of the commit"), N_("ISO-8601-TIMESTAMP") },
+#ifdef FLATPAK_ENABLE_P2P
+  { "collection-id", 0, 0, G_OPTION_ARG_STRING, &opt_collection_id, N_("Collection ID"), "COLLECTION-ID" },
+#endif  /* FLATPAK_ENABLE_P2P */
 
   { NULL }
 };
@@ -83,8 +91,20 @@ metadata_get_arch (GFile *file, char **out_arch, GError **error)
     return FALSE;
 
   runtime = g_key_file_get_string (keyfile,
-                                   opt_runtime ? "Runtime" : "Application",
+                                   "Application",
                                    "runtime", NULL);
+  if (runtime == NULL)
+    runtime = g_key_file_get_string (keyfile,
+                                     "Application",
+                                     "sdk", NULL);
+  if (runtime == NULL)
+    runtime = g_key_file_get_string (keyfile,
+                                     "Runtime",
+                                     "runtime", NULL);
+  if (runtime == NULL)
+    runtime = g_key_file_get_string (keyfile,
+                                     "Runtime",
+                                     "sdk", NULL);
   if (runtime == NULL)
     {
       *out_arch = g_strdup (flatpak_get_arch ());
@@ -154,11 +174,25 @@ commit_filter (OstreeRepo *repo,
   g_file_info_set_attribute_uint32 (file_info, "unix::uid", 0);
   g_file_info_set_attribute_uint32 (file_info, "unix::gid", 0);
 
+  /* In flatpak, there is no real reason for files to have different
+   * permissions based on the group or user really, everything is
+   * always used readonly for everyone. Having things be writeable
+   * for anyone but the user just causes risks for the system-installed
+   * case. So, we canonicalize the mode to writable only by the user,
+   * readable to all, and executable for all for directories and
+   * files that the user can execute.
+  */
   mode = g_file_info_get_attribute_uint32 (file_info, "unix::mode");
-  /* No setuid */
-  mode = mode & ~07000;
-  /* All files readable */
-  mode = mode | 0444;
+  if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_DIRECTORY)
+    mode = 0755 | S_IFDIR;
+  else if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_REGULAR)
+    {
+      /* If use can execute, make executable by all */
+      if (mode & S_IXUSR)
+        mode = 0755 | S_IFREG;
+      else /* otherwise executable by none */
+        mode = 0644 | S_IFREG;
+    }
   g_file_info_set_attribute_uint32 (file_info, "unix::mode", mode);
 
   if (matches_patterns (commit_data->exclude, path) &&
@@ -171,7 +205,7 @@ commit_filter (OstreeRepo *repo,
   return OSTREE_REPO_COMMIT_FILTER_ALLOW;
 }
 
-gboolean
+static gboolean
 add_file_to_mtree (GFile             *file,
                    const char        *name,
                    OstreeRepo        *repo,
@@ -304,7 +338,7 @@ validate_desktop_file (GFile *desktop_file,
   if (!subprocess)
     {
       if (!g_error_matches (local_error, G_SPAWN_ERROR, G_SPAWN_ERROR_NOENT))
-        g_print ("WARNING: Error running desktop-file-validate: %s\n", local_error->message);
+        g_print (_("WARNING: Error running desktop-file-validate: %s\n"), local_error->message);
 
       g_clear_error (&local_error);
       goto check_refs;
@@ -312,13 +346,13 @@ validate_desktop_file (GFile *desktop_file,
 
   if (!g_subprocess_communicate_utf8 (subprocess, NULL, NULL, &stdout_buf, &stderr_buf, &local_error))
     {
-      g_print ("WARNING: Error reading from desktop-file-validate: %s\n", local_error->message);
+      g_print (_("WARNING: Error reading from desktop-file-validate: %s\n"), local_error->message);
       g_clear_error (&local_error);
     }
 
   if (g_subprocess_get_if_exited (subprocess) &&
       g_subprocess_get_exit_status (subprocess) != 0)
-    g_print ("WARNING: Failed to validate desktop file %s: %s\n", path, stdout_buf);
+    g_print (_("WARNING: Failed to validate desktop file %s: %s\n"), path, stdout_buf);
 
 check_refs:
   /* Test that references to other files are valid */
@@ -333,7 +367,7 @@ check_refs:
                                    &local_error);
   if (!command)
     {
-      g_print ("WARNING: Can't find Exec key in %s: %s\n", path, local_error->message);
+      g_print (_("WARNING: Can't find Exec key in %s: %s\n"), path, local_error->message);
       g_clear_error (&local_error);
     }
 
@@ -341,14 +375,14 @@ check_refs:
 
   bin_file = convert_app_absolute_path (argv[0], files);
   if (!g_file_query_exists (bin_file, NULL))
-    g_print ("WARNING: Binary not found for Exec line in %s: %s\n", path, command);
+    g_print (_("WARNING: Binary not found for Exec line in %s: %s\n"), path, command);
 
   *icon = g_key_file_get_string (key_file,
                                  G_KEY_FILE_DESKTOP_GROUP,
                                  G_KEY_FILE_DESKTOP_KEY_ICON,
                                  NULL);
   if (*icon && !g_str_has_prefix (*icon, app_id))
-    g_print ("WARNING: Icon not matching app id in %s: %s\n", path, *icon);
+    g_print (_("WARNING: Icon not matching app id in %s: %s\n"), path, *icon);
 
   *activatable = g_key_file_get_boolean (key_file,
                                          G_KEY_FILE_DESKTOP_GROUP,
@@ -376,7 +410,7 @@ validate_icon (const char *icon,
   svg  = g_strconcat (icon, ".svg", NULL);
   if (!find_file_in_tree (icondir, png) &&
       !find_file_in_tree (icondir, svg))
-    g_print ("WARNING: Icon referenced in desktop file but not exported: %s\n", icon);
+    g_print (_("WARNING: Icon referenced in desktop file but not exported: %s\n"), icon);
 
   return TRUE;
 }
@@ -438,7 +472,7 @@ validate_service_file (GFile *service_file,
 
   bin_file = convert_app_absolute_path (argv[0], files);
   if (!g_file_query_exists (bin_file, NULL))
-    g_print ("WARNING: Binary not found for Exec line in %s: %s\n", path, command);
+    g_print (_("WARNING: Binary not found for Exec line in %s: %s\n"), path, command);
 
   return TRUE;
 }
@@ -601,7 +635,7 @@ flatpak_builtin_build_export (int argc, char **argv, GCancellable *cancellable,
   const char *branch;
   g_autofree char *arch = NULL;
   g_autofree char *full_branch = NULL;
-  g_autofree char *app_id = NULL;
+  g_autofree char *id = NULL;
   g_autofree char *parent = NULL;
   g_autofree char *commit_checksum = NULL;
   g_autofree char *metadata_contents = NULL;
@@ -619,6 +653,11 @@ flatpak_builtin_build_export (int argc, char **argv, GCancellable *cancellable,
   CommitData commit_data = {0};
   g_auto(GVariantDict) metadata_dict = FLATPAK_VARIANT_DICT_INITIALIZER;
   g_autoptr(GVariant) metadata_dict_v = NULL;
+  gboolean is_runtime = FALSE;
+  gboolean is_extension = FALSE;
+  guint64 installed_size = 0,download_size = 0;
+  GTimeVal ts;
+  const char *collection_id;
 
   context = g_option_context_new (_("LOCATION DIRECTORY [BRANCH] - Create a repository from a build directory"));
   g_option_context_set_translation_domain (context, GETTEXT_PACKAGE);
@@ -646,6 +685,15 @@ flatpak_builtin_build_export (int argc, char **argv, GCancellable *cancellable,
   else
     branch = "master";
 
+#ifdef FLATPAK_ENABLE_P2P
+  if (opt_collection_id != NULL &&
+      !ostree_validate_collection_id (opt_collection_id, &my_error))
+    {
+      flatpak_fail (error, _("‘%s’ is not a valid collection ID: %s"), opt_collection_id, my_error->message);
+      goto out;
+    }
+#endif  /* FLATPAK_ENABLE_P2P */
+
   if (!flatpak_is_valid_branch (branch, &my_error))
     {
       flatpak_fail (error, _("'%s' is not a valid branch name: %s"), branch, my_error->message);
@@ -681,11 +729,22 @@ flatpak_builtin_build_export (int argc, char **argv, GCancellable *cancellable,
   if (!g_key_file_load_from_data (metakey, metadata_contents, metadata_size, 0, error))
     goto out;
 
-  app_id = g_key_file_get_string (metakey, opt_runtime ? "Runtime" : "Application", "name", error);
-  if (app_id == NULL)
-    goto out;
+  id = g_key_file_get_string (metakey, "Application", "name", NULL);
+  if (id == NULL)
+    {
+      id = g_key_file_get_string (metakey, "Runtime", "name", NULL);
+      if (id == NULL)
+        {
+          flatpak_fail (error, _("No name specified in the metadata"));
+          goto out;
+        }
+      is_runtime = TRUE;
 
-  if (!opt_runtime && !g_file_query_exists (export, cancellable))
+      if (g_key_file_has_group (metakey, "ExtensionOf"))
+        is_extension = TRUE;
+    }
+
+  if (!(opt_runtime || is_runtime) && !g_file_query_exists (export, cancellable))
     {
       flatpak_fail (error, "Build directory %s not finalized, use flatpak build-finish", directory);
       goto out;
@@ -696,7 +755,8 @@ flatpak_builtin_build_export (int argc, char **argv, GCancellable *cancellable,
   if (!collect_extra_data (metakey, &metadata_dict, error))
     goto out;
 
-  if (!validate_exports (export, files, app_id, error))
+  if (!(opt_runtime || is_runtime) &&
+      !validate_exports (export, files, id, error))
     goto out;
 
   if (!metadata_get_arch (metadata, &arch, error))
@@ -705,13 +765,17 @@ flatpak_builtin_build_export (int argc, char **argv, GCancellable *cancellable,
   if (opt_subject)
     subject = g_strdup (opt_subject);
   else
-    subject = g_strconcat ("Export ", app_id, NULL);
+    subject = g_strconcat ("Export ", id, NULL);
   if (opt_body)
     body = g_strdup (opt_body);
   else
-    body = g_strconcat ("Name: ", app_id, "\nArch: ", arch, "\nBranch: ", branch, NULL);
+    body = g_strdup_printf ("Name: %s\n"
+                            "Arch: %s\n"
+                            "Branch: %s\n"
+                            "Built with: "PACKAGE_STRING"\n",
+                            id, arch, branch);
 
-  full_branch = g_strconcat (opt_runtime ? "runtime/" : "app/", app_id, "/", arch, "/", branch, NULL);
+  full_branch = g_strconcat ((opt_runtime || is_runtime) ? "runtime/" : "app/", id, "/", arch, "/", branch, NULL);
 
   repofile = g_file_new_for_commandline_arg (location);
   repo = ostree_repo_new (repofile);
@@ -724,16 +788,46 @@ flatpak_builtin_build_export (int argc, char **argv, GCancellable *cancellable,
 
       if (!ostree_repo_resolve_rev (repo, full_branch, TRUE, &parent, error))
         goto out;
+
+#ifdef FLATPAK_ENABLE_P2P
+      if (opt_collection_id != NULL &&
+          g_strcmp0 (ostree_repo_get_collection_id (repo), opt_collection_id) != 0)
+        {
+          flatpak_fail (error, "Specified collection ID ‘%s’ doesn’t match collection ID in repository configuration ‘%s’.",
+                        opt_collection_id, ostree_repo_get_collection_id (repo));
+          goto out;
+        }
+#endif  /* FLATPAK_ENABLE_P2P */
     }
   else
     {
+#ifdef FLATPAK_ENABLE_P2P
+      if (opt_collection_id != NULL &&
+          !ostree_repo_set_collection_id (repo, opt_collection_id, error))
+        goto out;
+#endif  /* FLATPAK_ENABLE_P2P */
       if (!ostree_repo_create (repo, OSTREE_REPO_MODE_ARCHIVE_Z2, cancellable, error))
         goto out;
     }
 
+  /* Get the canonical collection ID which we’ll use for the commit. This might
+   * be %NULL if the existing repo doesn’t have one and none was specified on
+   * the command line. */
+#ifdef FLATPAK_ENABLE_P2P
+  collection_id = ostree_repo_get_collection_id (repo);
+#else  /* if !FLATPAK_ENABLE_P2P */
+  collection_id = NULL;
+#endif  /* !FLATPAK_ENABLE_P2P */
+
   if (!ostree_repo_prepare_transaction (repo, NULL, cancellable, error))
     goto out;
 
+  /* This is useful only if the target is a "bare" rep, but this happens
+     in flatpak-builder when commiting to the cache repo. For other repos
+     this is a no-op */
+  if (!ostree_repo_scan_hardlinks (repo, cancellable, error))
+    goto out;
+
   mtree = ostree_mutable_tree_new ();
 
   if (!flatpak_mtree_create_root (repo, mtree, cancellable, error))
@@ -745,7 +839,16 @@ flatpak_builtin_build_export (int argc, char **argv, GCancellable *cancellable,
   modifier = ostree_repo_commit_modifier_new (OSTREE_REPO_COMMIT_MODIFIER_FLAGS_SKIP_XATTRS,
                                               (OstreeRepoCommitFilter) commit_filter, &commit_data, NULL);
 
-  if (opt_runtime)
+  if (is_extension)
+    {
+      commit_data.exclude = (const char **) opt_exclude;
+      commit_data.include = (const char **) opt_include;
+      if (!ostree_repo_write_directory_to_mtree (repo, files, files_mtree, modifier, cancellable, error))
+        goto out;
+      commit_data.exclude = NULL;
+      commit_data.include = NULL;
+    }
+  else if (opt_runtime || is_runtime)
     {
       commit_data.exclude = (const char **) opt_exclude;
       commit_data.include = (const char **) opt_include;
@@ -776,12 +879,52 @@ flatpak_builtin_build_export (int argc, char **argv, GCancellable *cancellable,
   if (!ostree_repo_write_mtree (repo, mtree, &root, cancellable, error))
     goto out;
 
-  metadata_dict_v = g_variant_ref_sink (g_variant_dict_end (&metadata_dict));
-  if (!ostree_repo_write_commit (repo, parent, subject, body, metadata_dict_v,
-                                 OSTREE_REPO_FILE (root),
-                                 &commit_checksum, cancellable, error))
+  if (!flatpak_repo_collect_sizes (repo, root, &installed_size, &download_size, cancellable, error))
     goto out;
 
+  /* Binding information. xa.ref is deprecated in favour of the OSTree keys, but
+   * keep it around for backwards compatibility. Write the bindings even if
+   * we’re compiled without P2P support, since other flatpak builds might be. */
+  g_variant_dict_insert_value (&metadata_dict, "ostree.collection-binding",
+                               g_variant_new_string ((collection_id != NULL) ? collection_id : ""));
+  g_variant_dict_insert_value (&metadata_dict, "ostree.ref-binding",
+                               g_variant_new_strv ((const gchar * const *) &full_branch, 1));
+  g_variant_dict_insert_value (&metadata_dict, "xa.ref", g_variant_new_string (full_branch));
+
+  g_variant_dict_insert_value (&metadata_dict, "xa.metadata", g_variant_new_string (metadata_contents));
+  g_variant_dict_insert_value (&metadata_dict, "xa.installed-size", g_variant_new_uint64 (GUINT64_TO_BE (installed_size)));
+  g_variant_dict_insert_value (&metadata_dict, "xa.download-size", g_variant_new_uint64 (GUINT64_TO_BE (download_size)));
+
+  metadata_dict_v = g_variant_ref_sink (g_variant_dict_end (&metadata_dict));
+
+  /* The timestamp is used for the commit metadata and AppStream data */
+  if (opt_timestamp != NULL)
+    {
+      if (!g_time_val_from_iso8601 (opt_timestamp, &ts))
+        {
+          g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                       "Could not parse '%s'", opt_timestamp);
+          goto out;
+        }
+    }
+
+  if (opt_timestamp == NULL)
+    {
+      if (!ostree_repo_write_commit (repo, parent, subject, body, metadata_dict_v,
+                                     OSTREE_REPO_FILE (root),
+                                     &commit_checksum, cancellable, error))
+        goto out;
+    }
+  else
+    {
+      if (!ostree_repo_write_commit_with_time (repo, parent, subject, body,
+                                               metadata_dict_v,
+                                               OSTREE_REPO_FILE (root),
+                                               ts.tv_sec, &commit_checksum,
+                                               cancellable, error))
+        goto out;
+    }
+
   if (opt_gpg_key_ids)
     {
       char **iter;
@@ -800,13 +943,24 @@ flatpak_builtin_build_export (int argc, char **argv, GCancellable *cancellable,
         }
     }
 
-  ostree_repo_transaction_set_ref (repo, NULL, full_branch, commit_checksum);
+#ifdef FLATPAK_ENABLE_P2P
+  if (collection_id != NULL)
+    {
+      OstreeCollectionRef ref = { (char *) collection_id, full_branch };
+      ostree_repo_transaction_set_collection_ref (repo, &ref, commit_checksum);
+    }
+  else
+#endif  /* FLATPAK_ENABLE_P2P */
+    {
+      ostree_repo_transaction_set_ref (repo, NULL, full_branch, commit_checksum);
+    }
 
   if (!ostree_repo_commit_transaction (repo, &stats, cancellable, error))
     goto out;
 
   if (opt_update_appstream &&
-      !flatpak_repo_generate_appstream (repo, (const char **) opt_gpg_key_ids, opt_gpg_homedir, cancellable, error))
+      !flatpak_repo_generate_appstream (repo, (const char **) opt_gpg_key_ids, opt_gpg_homedir,
+                                        (opt_timestamp != NULL) ? ts.tv_sec : 0, cancellable, error))
     return FALSE;
 
   if (!opt_no_update_summary &&
@@ -819,12 +973,13 @@ flatpak_builtin_build_export (int argc, char **argv, GCancellable *cancellable,
 
   format_size = g_format_size (stats.content_bytes_written);
 
-  g_print ("Commit: %s\n", commit_checksum);
-  g_print ("Metadata Total: %u\n", stats.metadata_objects_total);
-  g_print ("Metadata Written: %u\n", stats.metadata_objects_written);
-  g_print ("Content Total: %u\n", stats.content_objects_total);
-  g_print ("Content Written: %u\n", stats.content_objects_written);
-  g_print ("Content Bytes Written: %" G_GUINT64_FORMAT " (%s)\n", stats.content_bytes_written, format_size);
+  g_print (_("Commit: %s\n"), commit_checksum);
+  g_print (_("Metadata Total: %u\n"), stats.metadata_objects_total);
+  g_print (_("Metadata Written: %u\n"), stats.metadata_objects_written);
+  g_print (_("Content Total: %u\n"), stats.content_objects_total);
+  g_print (_("Content Written: %u\n"), stats.content_objects_written);
+  g_print (_("Content Bytes Written:"));
+  g_print (" %" G_GUINT64_FORMAT " (%s)\n", stats.content_bytes_written, format_size);
 
   ret = TRUE;
 
