@@ -37,14 +37,23 @@
 static char *opt_command;
 static char *opt_require_version;
 static char **opt_extra_data;
+static char **opt_extensions;
+static char **opt_metadata;
 static gboolean opt_no_exports;
+static int opt_extension_prio = G_MININT;
+static char *opt_sdk;
+static char *opt_runtime;
 
 static GOptionEntry options[] = {
   { "command", 0, 0, G_OPTION_ARG_STRING, &opt_command, N_("Command to set"), N_("COMMAND") },
   { "require-version", 0, 0, G_OPTION_ARG_STRING, &opt_require_version, N_("Flatpak version to require"), N_("MAJOR.MINOR.MICRO") },
   { "no-exports", 0, 0, G_OPTION_ARG_NONE, &opt_no_exports, N_("Don't process exports"), NULL },
-  { "extra-data", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_extra_data, N_("Extra data info"),
-    N_("NAME:SHA256:DOWNLOAD-SIZE:INSTALL-SIZE:URL") },
+  { "extra-data", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_extra_data, N_("Extra data info") },
+  { "extension", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_extensions, N_("Add extension point info"),  N_("NAME=VARIABLE[=VALUE]") },
+  { "extension-priority", 0, 0, G_OPTION_ARG_INT, &opt_extension_prio, N_("Set extension priority (only for extensions)"), N_("0") },
+  { "sdk", 0, 0, G_OPTION_ARG_STRING, &opt_sdk, N_("Change the sdk used for the app"),  N_("SDK") },
+  { "runtime", 0, 0, G_OPTION_ARG_STRING, &opt_runtime, N_("Change the runtime used for the app"),  N_("RUNTIME") },
+  { "metadata", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_metadata, N_("Set generic metadata option"),  N_("GROUP=KEY[=VALUE]") },
   { NULL }
 };
 
@@ -61,7 +70,7 @@ export_dir (int           source_parent_fd,
   int res;
 
   g_auto(GLnxDirFdIterator) source_iter = {0};
-  glnx_fd_close int destination_dfd = -1;
+  glnx_autofd int destination_dfd = -1;
   struct dirent *dent;
 
   if (!glnx_dirfd_iterator_init_at (source_parent_fd, source_name, FALSE, &source_iter, error))
@@ -127,11 +136,11 @@ export_dir (int           source_parent_fd,
 
           if (!flatpak_has_name_prefix (dent->d_name, required_prefix))
             {
-              g_print ("Not exporting %s, wrong prefix\n", source_printable);
+              g_print (_("Not exporting %s, wrong prefix\n"), source_printable);
               continue;
             }
 
-          g_print ("Exporting %s\n", source_printable);
+          g_print (_("Exporting %s\n"), source_printable);
 
           if (!glnx_file_copy_at (source_iter.fd, dent->d_name, &stbuf,
                                   destination_dfd, dent->d_name,
@@ -143,7 +152,7 @@ export_dir (int           source_parent_fd,
       else
         {
           source_printable = g_build_filename (source_relpath, dent->d_name, NULL);
-          g_debug ("Not exporting non-regular file %s\n", source_printable);
+          g_debug ("Not exporting non-regular file %s", source_printable);
         }
     }
 
@@ -185,11 +194,11 @@ copy_exports (GFile        *source,
 }
 
 static gboolean
-collect_exports (GFile *base, const char *app_id, gboolean is_runtime, GCancellable *cancellable, GError **error)
+collect_exports (GFile *base, const char *app_id, GCancellable *cancellable, GError **error)
 {
   g_autoptr(GFile) files = NULL;
   g_autoptr(GFile) export = NULL;
-  const char *app_paths[] = {
+  const char *paths[] = {
     "share/applications",                 /* Copy desktop files */
     "share/mime/packages",                /* Copy MIME Type files */
     "share/icons",                        /* Icons */
@@ -197,22 +206,9 @@ collect_exports (GFile *base, const char *app_id, gboolean is_runtime, GCancella
     "share/gnome-shell/search-providers", /* Search providers */
     NULL,
   };
-  const char *runtime_paths[] = {
-    NULL,
-  };
-  const char **paths;
   int i;
 
-  if (is_runtime)
-    {
-      files = g_file_get_child (base, "usr");
-      paths = runtime_paths;
-    }
-  else
-    {
-      files = g_file_get_child (base, "files");
-      paths = app_paths;
-    }
+  files = g_file_get_child (base, "files");
 
   export = g_file_get_child (base, "export");
 
@@ -247,7 +243,7 @@ collect_exports (GFile *base, const char *app_id, gboolean is_runtime, GCancella
 }
 
 static gboolean
-update_metadata (GFile *base, FlatpakContext *arg_context, GCancellable *cancellable, GError **error)
+update_metadata (GFile *base, FlatpakContext *arg_context, gboolean is_runtime, GCancellable *cancellable, GError **error)
 {
   gboolean ret = FALSE;
 
@@ -256,74 +252,158 @@ update_metadata (GFile *base, FlatpakContext *arg_context, GCancellable *cancell
   g_autoptr(GKeyFile) keyfile = NULL;
   g_autoptr(FlatpakContext) app_context = NULL;
   GError *temp_error = NULL;
+  const char *group;
   int i;
 
   metadata = g_file_get_child (base, "metadata");
   if (!g_file_query_exists (metadata, cancellable))
     goto out;
 
+  if (is_runtime)
+    group = FLATPAK_METADATA_GROUP_RUNTIME;
+  else
+    group = FLATPAK_METADATA_GROUP_APPLICATION;
+
   path = g_file_get_path (metadata);
   keyfile = g_key_file_new ();
   if (!g_key_file_load_from_file (keyfile, path, G_KEY_FILE_NONE, error))
     goto out;
 
-  if (g_key_file_has_key (keyfile, "Application", "command", NULL))
+  if (opt_sdk != NULL || opt_runtime != NULL)
     {
-      g_debug ("Command key is present");
+      g_autofree char *old_runtime = g_key_file_get_string (keyfile,
+                                                            group,
+                                                            FLATPAK_METADATA_KEY_RUNTIME, NULL);
+      g_autofree char *old_sdk = g_key_file_get_string (keyfile,
+                                                        group,
+                                                        FLATPAK_METADATA_KEY_SDK,
+                                                        NULL);
+      const char *old_sdk_arch = NULL;
+      const char *old_runtime_arch = NULL;
+      const char *old_sdk_branch = NULL;
+      const char *old_runtime_branch = NULL;
+      g_auto(GStrv) old_sdk_parts = NULL;
+      g_auto(GStrv) old_runtime_parts = NULL;
 
-      if (opt_command)
-        g_key_file_set_string (keyfile, "Application", "command", opt_command);
-    }
-  else if (opt_command)
-    {
-      g_debug ("Using explicitly provided command %s", opt_command);
+      if (old_sdk)
+        old_sdk_parts = g_strsplit (old_sdk, "/", 3);
 
-      g_key_file_set_string (keyfile, "Application", "command", opt_command);
-    }
-  else
-    {
-      g_autofree char *command = NULL;
-      g_autoptr(GFile) bin_dir = NULL;
-      g_autoptr(GFileEnumerator) bin_enum = NULL;
-      g_autoptr(GFileInfo) child_info = NULL;
+      if (old_runtime)
+        old_runtime_parts = g_strsplit (old_runtime, "/", 3);
 
-      g_debug ("Looking for executables");
-
-      bin_dir = g_file_resolve_relative_path (base, "files/bin");
-      if (g_file_query_exists (bin_dir, cancellable))
+      if (old_sdk_parts != NULL)
         {
-          bin_enum = g_file_enumerate_children (bin_dir, G_FILE_ATTRIBUTE_STANDARD_NAME,
-                                                G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
-                                                cancellable, error);
-          if (!bin_enum)
-            goto out;
+          old_sdk_arch = old_sdk_parts[1];
+          old_sdk_branch = old_sdk_parts[2];
+        }
+      if (old_runtime_parts != NULL)
+        {
+          old_runtime_arch = old_runtime_parts[1];
+          old_runtime_branch = old_runtime_parts[2];
 
-          while ((child_info = g_file_enumerator_next_file (bin_enum, cancellable, &temp_error)))
+          /* Use the runtime as fallback if sdk is not specified */
+          if (old_sdk_parts == NULL)
             {
-              if (command != NULL)
-                {
-                  g_print ("More than one executable found\n");
-                  break;
-                }
-              command = g_strdup (g_file_info_get_name (child_info));
+              old_sdk_arch = old_runtime_parts[1];
+              old_sdk_branch = old_runtime_parts[2];
             }
-          if (temp_error != NULL)
-            goto out;
         }
 
-      if (command)
+      if (opt_sdk)
         {
-          g_print ("Using %s as command\n", command);
-          g_key_file_set_string (keyfile, "Application", "command", command);
+          g_autofree char *id = NULL;
+          g_autofree char *arch = NULL;
+          g_autofree char *branch = NULL;
+          g_autofree char *ref = NULL;
+          FlatpakKinds kinds;
+
+          if (!flatpak_split_partial_ref_arg (opt_sdk, FLATPAK_KINDS_RUNTIME,
+                                              old_sdk_arch, old_sdk_branch ? old_sdk_branch : "master",
+                                              &kinds, &id, &arch, &branch, error))
+            return FALSE;
+
+          ref = flatpak_build_untyped_ref (id, branch, arch);
+          g_key_file_set_string (keyfile, group, FLATPAK_METADATA_KEY_SDK, ref);
+        }
+
+      if (opt_runtime)
+        {
+          g_autofree char *id = NULL;
+          g_autofree char *arch = NULL;
+          g_autofree char *branch = NULL;
+          g_autofree char *ref = NULL;
+          FlatpakKinds kinds;
+
+          if (!flatpak_split_partial_ref_arg (opt_runtime, FLATPAK_KINDS_RUNTIME,
+                                              old_runtime_arch, old_runtime_branch ? old_runtime_branch : "master",
+                                              &kinds, &id, &arch, &branch, error))
+            return FALSE;
+
+          ref = flatpak_build_untyped_ref (id, branch, arch);
+          g_key_file_set_string (keyfile, group, FLATPAK_METADATA_KEY_RUNTIME, ref);
+        }
+    }
+
+  if (!is_runtime)
+    {
+      if (g_key_file_has_key (keyfile, group, FLATPAK_METADATA_KEY_COMMAND, NULL))
+        {
+          g_debug ("Command key is present");
+
+          if (opt_command)
+            g_key_file_set_string (keyfile, group, FLATPAK_METADATA_KEY_COMMAND, opt_command);
+        }
+      else if (opt_command)
+        {
+          g_debug ("Using explicitly provided command %s", opt_command);
+
+          g_key_file_set_string (keyfile, group, FLATPAK_METADATA_KEY_COMMAND, opt_command);
         }
       else
         {
-          g_print ("No executable found\n");
+          g_autofree char *command = NULL;
+          g_autoptr(GFile) bin_dir = NULL;
+          g_autoptr(GFileEnumerator) bin_enum = NULL;
+          g_autoptr(GFileInfo) child_info = NULL;
+
+          g_debug ("Looking for executables");
+
+          bin_dir = g_file_resolve_relative_path (base, "files/bin");
+          if (g_file_query_exists (bin_dir, cancellable))
+            {
+              bin_enum = g_file_enumerate_children (bin_dir, G_FILE_ATTRIBUTE_STANDARD_NAME,
+                                                    G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
+                                                    cancellable, error);
+              if (!bin_enum)
+                goto out;
+
+              while ((child_info = g_file_enumerator_next_file (bin_enum, cancellable, &temp_error)))
+                {
+                  if (command != NULL)
+                    {
+                      g_print (_("More than one executable found\n"));
+                      break;
+                    }
+                  command = g_strdup (g_file_info_get_name (child_info));
+                }
+              if (temp_error != NULL)
+                goto out;
+            }
+
+          if (command)
+            {
+              g_print (_("Using %s as command\n"), command);
+              g_key_file_set_string (keyfile, group, FLATPAK_METADATA_KEY_COMMAND, command);
+            }
+          else
+            {
+              g_print (_("No executable found\n"));
+            }
         }
     }
 
   if (opt_require_version)
-    g_key_file_set_string (keyfile, "Application", "required-flatpak", opt_require_version);
+    g_key_file_set_string (keyfile, group, "required-flatpak", opt_require_version);
 
   app_context = flatpak_context_new ();
   if (!flatpak_context_load_metadata (app_context, keyfile, error))
@@ -350,25 +430,68 @@ update_metadata (GFile *base, FlatpakContext *arg_context, GCancellable *cancell
       elements = g_strsplit (extra_data, ":", 5);
       if (g_strv_length (elements) != 5)
         {
-          flatpak_fail (error, _("To few elements in --extra-data argument %s"), extra_data);
+          flatpak_fail (error, _("Too few elements in --extra-data argument %s"), extra_data);
           goto out;
         }
 
-      uri_key = g_strconcat ("uri", suffix, NULL);
-      name_key = g_strconcat ("name", suffix, NULL);
-      checksum_key = g_strconcat ("checksum", suffix, NULL);
-      size_key = g_strconcat ("size", suffix, NULL);
-      installed_size_key = g_strconcat ("installed-size", suffix, NULL);
+      uri_key = g_strconcat (FLATPAK_METADATA_KEY_EXTRA_DATA_URI, suffix, NULL);
+      name_key = g_strconcat (FLATPAK_METADATA_KEY_EXTRA_DATA_NAME, suffix, NULL);
+      checksum_key = g_strconcat (FLATPAK_METADATA_KEY_EXTRA_DATA_CHECKSUM,
+                                  suffix, NULL);
+      size_key = g_strconcat (FLATPAK_METADATA_KEY_EXTRA_DATA_SIZE, suffix, NULL);
+      installed_size_key = g_strconcat (FLATPAK_METADATA_KEY_EXTRA_DATA_INSTALLED_SIZE,
+                                        suffix, NULL);
 
       if (strlen (elements[0]) > 0)
-        g_key_file_set_string (keyfile, "Extra Data", name_key, elements[0]);
-      g_key_file_set_string (keyfile, "Extra Data", checksum_key, elements[1]);
-      g_key_file_set_string (keyfile, "Extra Data", size_key, elements[2]);
+        g_key_file_set_string (keyfile, FLATPAK_METADATA_GROUP_EXTRA_DATA,
+                               name_key, elements[0]);
+      g_key_file_set_string (keyfile, FLATPAK_METADATA_GROUP_EXTRA_DATA,
+                             checksum_key, elements[1]);
+      g_key_file_set_string (keyfile, FLATPAK_METADATA_GROUP_EXTRA_DATA,
+                             size_key, elements[2]);
       if (strlen (elements[3]) > 0)
-        g_key_file_set_string (keyfile, "Extra Data", installed_size_key, elements[3]);
-      g_key_file_set_string (keyfile, "Extra Data", uri_key, elements[4]);
+        g_key_file_set_string (keyfile, FLATPAK_METADATA_GROUP_EXTRA_DATA,
+                               installed_size_key, elements[3]);
+      g_key_file_set_string (keyfile, FLATPAK_METADATA_GROUP_EXTRA_DATA,
+                             uri_key, elements[4]);
     }
 
+  for (i = 0; opt_metadata != NULL && opt_metadata[i] != NULL; i++)
+    {
+      g_auto(GStrv) elements = NULL;
+      elements = g_strsplit (opt_metadata[i], "=", 3);
+      if (g_strv_length (elements) < 2)
+        {
+          flatpak_fail (error, _("Too few elements in --metadata argument %s, format should be GROUP=KEY[=VALUE]]"), opt_metadata[i]);
+          goto out;
+        }
+
+      g_key_file_set_string (keyfile, elements[0], elements[1], elements[2] ? elements[2] : "true");
+    }
+
+  for (i = 0; opt_extensions != NULL && opt_extensions[i] != NULL; i++)
+    {
+      g_auto(GStrv) elements = NULL;
+      g_autofree char *groupname = NULL;
+
+      elements = g_strsplit (opt_extensions[i], "=", 3);
+      if (g_strv_length (elements) < 2)
+        {
+          flatpak_fail (error, _("Too few elements in --extension argument %s, format should be NAME=VAR[=VALUE]"), opt_extensions[i]);
+          goto out;
+        }
+
+      groupname = g_strconcat (FLATPAK_METADATA_GROUP_PREFIX_EXTENSION,
+                               elements[0], NULL);
+
+      g_key_file_set_string (keyfile, groupname, elements[1], elements[2] ? elements[2] : "true");
+    }
+
+
+  if (opt_extension_prio != G_MININT)
+    g_key_file_set_integer (keyfile, FLATPAK_METADATA_GROUP_EXTENSION_OF,
+                            FLATPAK_METADATA_KEY_PRIORITY, opt_extension_prio);
+
   if (!g_key_file_save_to_file (keyfile, path, error))
     goto out;
 
@@ -428,10 +551,12 @@ flatpak_builtin_build_finish (int argc, char **argv, GCancellable *cancellable,
   if (!g_key_file_load_from_data (metakey, metadata_contents, metadata_size, 0, error))
     return FALSE;
 
-  id = g_key_file_get_string (metakey, "Application", "name", NULL);
+  id = g_key_file_get_string (metakey, FLATPAK_METADATA_GROUP_APPLICATION,
+                              FLATPAK_METADATA_KEY_NAME, NULL);
   if (id == NULL)
     {
-      id = g_key_file_get_string (metakey, "Runtime", "name", NULL);
+      id = g_key_file_get_string (metakey, FLATPAK_METADATA_GROUP_RUNTIME,
+                                  FLATPAK_METADATA_KEY_NAME, NULL);
       if (id == NULL)
         return flatpak_fail (error, _("No name specified in the metadata"));
       is_runtime = TRUE;
@@ -440,12 +565,15 @@ flatpak_builtin_build_finish (int argc, char **argv, GCancellable *cancellable,
   if (g_file_query_exists (export_dir, cancellable))
     return flatpak_fail (error, _("Build directory %s already finalized"), directory);
 
-  g_debug ("Collecting exports");
-  if (!collect_exports (base, id, is_runtime, cancellable, error))
-    return FALSE;
+  if (!is_runtime)
+    {
+      g_debug ("Collecting exports");
+      if (!collect_exports (base, id, cancellable, error))
+        return FALSE;
+    }
 
   g_debug ("Updating metadata");
-  if (!update_metadata (base, arg_context, cancellable, error))
+  if (!update_metadata (base, arg_context, is_runtime, cancellable, error))
     return FALSE;
 
   g_print (_("Please review the exported files and the metadata\n"));
