@@ -30,10 +30,12 @@
 #include "libglnx/libglnx.h"
 
 #include "flatpak-builtins.h"
+#include "flatpak-builtins-utils.h"
 #include "flatpak-utils.h"
 
 static char *opt_arch;
 static char *opt_var;
+static char *opt_type;
 static char *opt_sdk_dir;
 static char **opt_sdk_extensions;
 static char **opt_tags;
@@ -50,6 +52,7 @@ static GOptionEntry options[] = {
   { "base-version", 0, 0, G_OPTION_ARG_STRING, &opt_base_version, N_("Specify version for --base"), N_("VERSION") },
   { "base-extension", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_base_extensions, N_("Include this base extension"), N_("EXTENSION") },
   { "writable-sdk", 'w', 0, G_OPTION_ARG_NONE, &opt_writable_sdk, N_("Initialize /usr with a writable copy of the sdk"), NULL },
+  { "type", 0, 0, G_OPTION_ARG_STRING, &opt_type, N_("Specify the build type (app, runtime, extension)"), N_("TYPE") },
   { "tag", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_tags, N_("Add a tag"), N_("TAG") },
   { "sdk-extension", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_sdk_extensions, N_("Include this sdk extension in /usr"), N_("EXTENSION") },
   { "sdk-dir", 0, 0, G_OPTION_ARG_STRING, &opt_sdk_dir, N_("Where to store sdk (defaults to 'usr')"), N_("DIR") },
@@ -58,7 +61,7 @@ static GOptionEntry options[] = {
 };
 
 static gboolean
-copy_extensions (FlatpakDeploy *src_deploy, const char *default_branch,
+ensure_extensions (FlatpakDeploy *src_deploy, const char *default_branch,
                  char *src_extensions[], GFile *top_dir, GCancellable *cancellable, GError **error)
 {
   g_autoptr(GKeyFile) metakey = flatpak_deploy_get_metadata (src_deploy);
@@ -80,10 +83,6 @@ copy_extensions (FlatpakDeploy *src_deploy, const char *default_branch,
           if (strcmp (ext->installed_id, requested_extension) == 0 ||
               strcmp (ext->id, requested_extension) == 0)
             {
-              g_autoptr(GFile) target = g_file_resolve_relative_path (top_dir, ext->directory);
-              g_autoptr(GFile) target_parent = g_file_get_parent (target);
-              g_autoptr(GFile) ext_deploy_files = g_file_new_for_path (ext->files_path);
-
               if (!ext->is_unmaintained)
                 {
                   g_autoptr(FlatpakDir) src_dir = NULL;
@@ -103,17 +102,24 @@ copy_extensions (FlatpakDeploy *src_deploy, const char *default_branch,
                     return flatpak_fail (error, _("Requested extension %s is only partially installed"), ext->installed_id);
                 }
 
-              if (!flatpak_mkdir_p (target_parent, cancellable, error))
-                return FALSE;
+              if (top_dir)
+                {
+                  g_autoptr(GFile) target = g_file_resolve_relative_path (top_dir, ext->directory);
+                  g_autoptr(GFile) target_parent = g_file_get_parent (target);
+                  g_autoptr(GFile) ext_deploy_files = g_file_new_for_path (ext->files_path);
 
-              /* An extension overrides whatever is there before, so we clean up first */
-              if (!flatpak_rm_rf (target, cancellable, error))
-                return FALSE;
+                  if (!flatpak_mkdir_p (target_parent, cancellable, error))
+                    return FALSE;
 
-              if (!flatpak_cp_a (ext_deploy_files, target,
-                                 FLATPAK_CP_FLAGS_NO_CHOWN,
-                                 cancellable, error))
-                return FALSE;
+                  /* An extension overrides whatever is there before, so we clean up first */
+                  if (!flatpak_rm_rf (target, cancellable, error))
+                    return FALSE;
+
+                  if (!flatpak_cp_a (ext_deploy_files, target,
+                                     FLATPAK_CP_FLAGS_NO_CHOWN,
+                                     cancellable, error))
+                    return FALSE;
+                }
 
               found = TRUE;
             }
@@ -145,15 +151,22 @@ flatpak_builtin_build_init (int argc, char **argv, GCancellable *cancellable, GE
   g_autoptr(GFile) metadata_file = NULL;
   g_autoptr(GString) metadata_contents = NULL;
   g_autoptr(GError) my_error = NULL;
+  g_autoptr(FlatpakDeploy) sdk_deploy = NULL;
   const char *app_id;
   const char *directory;
-  const char *sdk;
-  const char *runtime;
-  const char *branch = "master";
+  const char *sdk_pref;
+  const char *runtime_pref;
+  const char *default_branch = NULL;
   g_autofree char *runtime_ref = NULL;
   g_autofree char *var_ref = NULL;
   g_autofree char *sdk_ref = NULL;
+  FlatpakKinds kinds;
   int i;
+  g_autoptr(FlatpakDir) sdk_dir = NULL;
+  g_autoptr(FlatpakDir) runtime_dir = NULL;
+  gboolean is_app = FALSE;
+  gboolean is_extension = FALSE;
+  gboolean is_runtime = FALSE;
 
   context = g_option_context_new (_("DIRECTORY APPNAME SDK RUNTIME [BRANCH] - Initialize a directory for building"));
   g_option_context_set_translation_domain (context, GETTEXT_PACKAGE);
@@ -169,25 +182,42 @@ flatpak_builtin_build_init (int argc, char **argv, GCancellable *cancellable, GE
 
   directory = argv[1];
   app_id = argv[2];
-  sdk = argv[3];
-  runtime = argv[4];
+  sdk_pref = argv[3];
+  runtime_pref = argv[4];
   if (argc >= 6)
-    branch = argv[5];
+    default_branch = argv[5];
+
+  if (opt_type != NULL)
+    {
+      if (strcmp (opt_type, "app") == 0)
+        is_app = TRUE;
+      else if (strcmp (opt_type, "extension") == 0)
+        is_extension = TRUE;
+      else if (strcmp (opt_type, "runtime") == 0)
+        is_runtime = TRUE;
+      else
+        return flatpak_fail (error, _("'%s' is not a valid build type name, use app, runtime or extension"), opt_type);
+    }
+  else
+    is_app = TRUE;
 
   if (!flatpak_is_valid_name (app_id, &my_error))
     return flatpak_fail (error, _("'%s' is not a valid application name: %s"), app_id, my_error->message);
 
-  if (!flatpak_is_valid_name (runtime, &my_error))
-    return flatpak_fail (error, _("'%s' is not a valid runtime name: %s"), runtime, my_error->message);
+  kinds = FLATPAK_KINDS_RUNTIME;
+  sdk_dir = flatpak_find_installed_pref (sdk_pref, kinds, opt_arch, default_branch, TRUE, FALSE, FALSE, NULL,
+                                         &sdk_ref, cancellable, error);
+  if (sdk_dir == NULL)
+    return FALSE;
 
-  if (!flatpak_is_valid_name (sdk, &my_error))
-    return flatpak_fail (error, _("'%s' is not a valid sdk name: %s"), sdk, my_error->message);
+  kinds = FLATPAK_KINDS_RUNTIME;
+  if (is_extension)
+    kinds |= FLATPAK_KINDS_APP;
 
-  if (!flatpak_is_valid_branch (branch, &my_error))
-    return flatpak_fail (error, _("'%s' is not a valid branch name: %s"), branch, my_error->message);
-
-  runtime_ref = flatpak_build_untyped_ref (runtime, branch, opt_arch);
-  sdk_ref = flatpak_build_untyped_ref (sdk, branch, opt_arch);
+  runtime_dir = flatpak_find_installed_pref (runtime_pref, kinds, opt_arch, default_branch, TRUE, FALSE, FALSE, NULL,
+                                             &runtime_ref, cancellable, error);
+  if (runtime_dir == NULL)
+    return FALSE;
 
   base = g_file_new_for_commandline_arg (directory);
 
@@ -195,10 +225,6 @@ flatpak_builtin_build_init (int argc, char **argv, GCancellable *cancellable, GE
     return FALSE;
 
   files_dir = g_file_get_child (base, "files");
-  if (opt_sdk_dir)
-    usr_dir = g_file_get_child (base, opt_sdk_dir);
-  else
-    usr_dir = g_file_get_child (base, "usr");
   var_dir = g_file_get_child (base, "var");
   var_tmp_dir = g_file_get_child (var_dir, "tmp");
   var_run_dir = g_file_get_child (var_dir, "run");
@@ -208,15 +234,18 @@ flatpak_builtin_build_init (int argc, char **argv, GCancellable *cancellable, GE
       g_file_query_exists (files_dir, cancellable))
     return flatpak_fail (error, _("Build directory %s already initialized"), directory);
 
-  if (opt_writable_sdk)
-    {
-      g_autofree char *full_sdk_ref = g_strconcat ("runtime/", sdk_ref, NULL);
-      g_autoptr(GFile) sdk_deploy_files = NULL;
-      g_autoptr(FlatpakDeploy) sdk_deploy = NULL;
+  sdk_deploy = flatpak_dir_load_deployed (sdk_dir, sdk_ref, NULL, cancellable, error);
+  if (sdk_deploy == NULL)
+    return FALSE;
 
-      sdk_deploy = flatpak_find_deploy_for_ref (full_sdk_ref, cancellable, error);
-      if (sdk_deploy == NULL)
-        return FALSE;
+  if (opt_writable_sdk || is_runtime)
+    {
+      g_autoptr(GFile) sdk_deploy_files = NULL;
+
+      if (opt_sdk_dir)
+        usr_dir = g_file_get_child (base, opt_sdk_dir);
+      else
+        usr_dir = g_file_get_child (base, "usr");
 
       if (!flatpak_rm_rf (usr_dir, NULL, &my_error))
         {
@@ -232,15 +261,16 @@ flatpak_builtin_build_init (int argc, char **argv, GCancellable *cancellable, GE
       sdk_deploy_files = flatpak_deploy_get_files (sdk_deploy);
       if (!flatpak_cp_a (sdk_deploy_files, usr_dir, FLATPAK_CP_FLAGS_NO_CHOWN, cancellable, error))
         return FALSE;
-
-      if (opt_sdk_extensions &&
-          !copy_extensions (sdk_deploy, branch, opt_sdk_extensions, usr_dir, cancellable, error))
-        return FALSE;
     }
 
+  if (opt_sdk_extensions &&
+      !ensure_extensions (sdk_deploy, default_branch,
+                          opt_sdk_extensions, usr_dir, cancellable, error))
+    return FALSE;
+
   if (opt_var)
     {
-      var_ref = flatpak_build_runtime_ref (opt_var, branch, opt_arch);
+      var_ref = flatpak_build_runtime_ref (opt_var, default_branch, opt_arch);
 
       var_deploy_files = flatpak_find_files_dir_for_ref (var_ref, cancellable, error);
       if (var_deploy_files == NULL)
@@ -274,7 +304,7 @@ flatpak_builtin_build_init (int argc, char **argv, GCancellable *cancellable, GE
 
 
       if (opt_base_extensions &&
-          !copy_extensions (base_deploy, base_branch, opt_base_extensions, files_dir, cancellable, error))
+          !ensure_extensions (base_deploy, base_branch, opt_base_extensions, files_dir, cancellable, error))
         return FALSE;
     }
 
@@ -297,12 +327,27 @@ flatpak_builtin_build_init (int argc, char **argv, GCancellable *cancellable, GE
     return FALSE;
 
 
-  metadata_contents = g_string_new ("[Application]\n");
+  metadata_contents = g_string_new ("");
+  if (is_app)
+    g_string_append (metadata_contents, "[Application]\n");
+  else
+    g_string_append (metadata_contents, "[Runtime]\n");
+
   g_string_append_printf (metadata_contents,
-                          "name=%s\n"
-                          "runtime=%s\n"
-                          "sdk=%s\n",
-                          app_id, runtime_ref, sdk_ref);
+                          "name=%s\n",
+                          app_id);
+
+  /* The "runtime" can be an app in case we're building an extension */
+  if (g_str_has_prefix (runtime_ref, "runtime/"))
+    g_string_append_printf (metadata_contents,
+                            "runtime=%s\n",
+                            runtime_ref + strlen ("runtime/"));
+
+  if (g_str_has_prefix (sdk_ref, "runtime/"))
+    g_string_append_printf (metadata_contents,
+                            "sdk=%s\n",
+                            sdk_ref + strlen ("runtime/"));
+
   if (opt_tags != NULL)
     {
       g_string_append (metadata_contents, "tags=");
@@ -314,6 +359,13 @@ flatpak_builtin_build_init (int argc, char **argv, GCancellable *cancellable, GE
       g_string_append_c (metadata_contents, '\n');
     }
 
+  if (is_extension)
+    g_string_append_printf (metadata_contents,
+                            "\n"
+                            "[ExtensionOf]\n"
+                            "ref=%s\n",
+                            runtime_ref);
+
   if (!g_file_replace_contents (metadata_file,
                                 metadata_contents->str, metadata_contents->len, NULL, FALSE,
                                 G_FILE_CREATE_REPLACE_DESTINATION,
