@@ -29,7 +29,7 @@
 #include <glib/gi18n.h>
 
 #include "libglnx/libglnx.h"
-#include "document-portal/xdp-dbus.h"
+#include "flatpak-document-dbus.h"
 
 #include <gio/gunixfdlist.h>
 
@@ -112,7 +112,7 @@ flatpak_complete_document_unexport (FlatpakCompletion *completion)
       flatpak_complete_options (completion, global_entries);
       flatpak_complete_options (completion, options);
 
-      flatpak_complete_file (completion);
+      flatpak_complete_file (completion, "__FLATPAK_FILE");
       break;
     }
 
