@@ -31,23 +31,20 @@
 
 #include "flatpak-builtins.h"
 #include "flatpak-utils.h"
+#include "flatpak-table-printer.h"
 
 static gboolean opt_show_details;
-static gboolean opt_user;
-static gboolean opt_system;
 static gboolean opt_runtime;
 static gboolean opt_app;
-static char **opt_installations;
+static gboolean opt_all;
 static char *opt_arch;
 
 static GOptionEntry options[] = {
-  { "user", 0, 0, G_OPTION_ARG_NONE, &opt_user, N_("Show user installations"), NULL },
-  { "system", 0, 0, G_OPTION_ARG_NONE, &opt_system, N_("Show system-wide installations"), NULL },
-  { "installation", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_installations, N_("Show specific system-wide installations"), NULL },
-  { "show-details", 'd', 0, G_OPTION_ARG_NONE, &opt_show_details, N_("Show arches and branches"), NULL },
+  { "show-details", 'd', 0, G_OPTION_ARG_NONE, &opt_show_details, N_("Show extra information"), NULL },
   { "runtime", 0, 0, G_OPTION_ARG_NONE, &opt_runtime, N_("List installed runtimes"), NULL },
   { "app", 0, 0, G_OPTION_ARG_NONE, &opt_app, N_("List installed applications"), NULL },
   { "arch", 0, 0, G_OPTION_ARG_STRING, &opt_arch, N_("Arch to show"), N_("ARCH") },
+  { "all", 'a', 0, G_OPTION_ARG_NONE, &opt_all, N_("List all refs (including locale/debug)"), NULL },
   { NULL }
 };
 
@@ -124,24 +121,48 @@ print_table_for_refs (gboolean print_apps, GPtrArray* refs_array, const char *ar
   FlatpakTablePrinter *printer = flatpak_table_printer_new ();
   int i;
 
+  i = 0;
+  flatpak_table_printer_set_column_title (printer, i++, _("Ref"));
+  if (opt_show_details)
+    {
+      flatpak_table_printer_set_column_title (printer, i++, _("Origin"));
+      flatpak_table_printer_set_column_title (printer, i++, _("Active commit"));
+      flatpak_table_printer_set_column_title (printer, i++, _("Latest commit"));
+      flatpak_table_printer_set_column_title (printer, i++, _("Installed size"));
+    }
+  flatpak_table_printer_set_column_title (printer, i++, _("Options"));
+
   for (i = 0; i < refs_array->len; i++)
     {
       RefsData *refs_data = NULL;
       FlatpakDir *dir = NULL;
       g_auto(GStrv) dir_refs = NULL;
-      g_autofree char *last = NULL;
+      g_autoptr(GHashTable) pref_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
       int j;
 
       refs_data = (RefsData *) g_ptr_array_index (refs_array, i);
       dir = refs_data->dir;
       dir_refs = join_strv (refs_data->app_refs, refs_data->runtime_refs);
 
+      for (j = 0; dir_refs[j] != NULL; j++)
+        {
+          char *ref = dir_refs[j];
+          char *partial_ref = flatpak_make_valid_id_prefix (strchr (ref, '/') + 1);
+          g_hash_table_insert (pref_hash, partial_ref, ref);
+        }
+
       for (j = 0; dir_refs[j] != NULL; j++)
         {
           char *ref, *partial_ref;
           g_auto(GStrv) parts = NULL;
           const char *repo = NULL;
           g_autoptr(GVariant) deploy_data = NULL;
+          const char *active;
+          const char *alt_id;
+          g_autofree char *latest = NULL;
+          g_autofree char *size_s = NULL;
+          guint64 size = 0;
+          g_autofree const char **subpaths = NULL;
 
           ref = dir_refs[j];
 
@@ -155,36 +176,48 @@ print_table_for_refs (gboolean print_apps, GPtrArray* refs_array, const char *ar
           if (deploy_data == NULL)
             continue;
 
+          if (!opt_all && strcmp (parts[0], "runtime") == 0 &&
+              flatpak_id_has_subref_suffix (parts[1]))
+            {
+              g_autofree char *prefix_partial_ref = NULL;
+              char *last_dot = strrchr (parts[1], '.');
+
+              *last_dot = 0;
+              prefix_partial_ref = g_strconcat (parts[1], "/", parts[2], "/", parts[3], NULL);
+              *last_dot = '.';
+
+              if (g_hash_table_lookup (pref_hash, prefix_partial_ref))
+                continue;
+            }
+
+
           repo = flatpak_deploy_data_get_origin (deploy_data);
 
-          if (opt_show_details)
-            {
-              const char *active = flatpak_deploy_data_get_commit (deploy_data);
-              const char *alt_id = flatpak_deploy_data_get_alt_id (deploy_data);
-              g_autofree char *latest = NULL;
-              g_autofree char *size_s = NULL;
-              guint64 size = 0;
-              g_autofree const char **subpaths = NULL;
+          active = flatpak_deploy_data_get_commit (deploy_data);
+          alt_id = flatpak_deploy_data_get_alt_id (deploy_data);
 
-              latest = flatpak_dir_read_latest (dir, repo, ref, NULL, NULL, NULL);
-              if (latest)
+          latest = flatpak_dir_read_latest (dir, repo, ref, NULL, NULL, NULL);
+          if (latest)
+            {
+              if (strcmp (active, latest) == 0)
                 {
-                  if (strcmp (active, latest) == 0)
-                    {
-                      g_free (latest);
-                      latest = g_strdup ("-");
-                    }
-                  else
-                    {
-                      latest[MIN (strlen (latest), 12)] = 0;
-                    }
+                  g_free (latest);
+                  latest = g_strdup ("-");
                 }
               else
                 {
-                  latest = g_strdup ("?");
+                  latest[MIN (strlen (latest), 12)] = 0;
                 }
+            }
+          else
+            {
+              latest = g_strdup ("?");
+            }
 
-              flatpak_table_printer_add_column (printer, partial_ref);
+          flatpak_table_printer_add_column (printer, partial_ref);
+
+          if (opt_show_details)
+            {
               flatpak_table_printer_add_column (printer, repo);
 
               flatpak_table_printer_add_column_len (printer, active, 12);
@@ -192,57 +225,41 @@ print_table_for_refs (gboolean print_apps, GPtrArray* refs_array, const char *ar
 
               size = flatpak_deploy_data_get_installed_size (deploy_data);
               size_s = g_format_size (size);
-              flatpak_table_printer_add_column (printer, size_s);
+              flatpak_table_printer_add_decimal_column (printer, size_s);
+            }
 
-              flatpak_table_printer_add_column (printer, ""); /* Options */
+          flatpak_table_printer_add_column (printer, ""); /* Options */
 
-              if (refs_array->len > 1)
-                {
-                  g_autofree char *source = flatpak_dir_get_name (dir);
-                  flatpak_table_printer_append_with_comma (printer, source);
-                }
+          if (refs_array->len > 1)
+            {
+              g_autofree char *source = flatpak_dir_get_name (dir);
+              flatpak_table_printer_append_with_comma (printer, source);
+            }
 
-              if (alt_id)
-                {
-                  g_autofree char *alt_id_str = g_strdup_printf ("alt-id=%.12s", alt_id);
-                  flatpak_table_printer_append_with_comma (printer, alt_id_str);
-                }
+          if (alt_id)
+            {
+              g_autofree char *alt_id_str = g_strdup_printf ("alt-id=%.12s", alt_id);
+              flatpak_table_printer_append_with_comma (printer, alt_id_str);
+            }
 
-              if (strcmp (parts[0], "app") == 0)
-                {
-                  g_autofree char *current = NULL;
+          if (strcmp (parts[0], "app") == 0)
+            {
+              g_autofree char *current = NULL;
 
-                  current = flatpak_dir_current_ref (dir, parts[1], cancellable);
-                  if (current && strcmp (ref, current) == 0)
-                    flatpak_table_printer_append_with_comma (printer, "current");
-                }
-              else
-                {
-                  if (print_apps)
-                    flatpak_table_printer_append_with_comma (printer, "runtime");
-                }
-
-              subpaths = flatpak_deploy_data_get_subpaths (deploy_data);
-              if (subpaths[0] == NULL)
-                {
-                  flatpak_table_printer_add_column (printer, "");
-                }
-              else
-                {
-                  int i;
-                  flatpak_table_printer_add_column (printer, ""); /* subpaths */
-                  for (i = 0; subpaths[i] != NULL; i++)
-                    flatpak_table_printer_append_with_comma (printer, subpaths[i]);
-                }
+              current = flatpak_dir_current_ref (dir, parts[1], cancellable);
+              if (current && strcmp (ref, current) == 0)
+                flatpak_table_printer_append_with_comma (printer, "current");
             }
           else
             {
-              if (last == NULL || strcmp (last, parts[1]) != 0)
-                {
-                  flatpak_table_printer_add_column (printer, parts[1]);
-                  g_clear_pointer (&last, g_free);
-                  last = g_strdup (parts[1]);
-                }
+              if (print_apps)
+                flatpak_table_printer_append_with_comma (printer, "runtime");
+            }
+
+          subpaths = flatpak_deploy_data_get_subpaths (deploy_data);
+          if (subpaths[0] != NULL)
+            {
+              flatpak_table_printer_append_with_comma (printer, "partial");
             }
           flatpak_table_printer_finish_row (printer);
         }
@@ -255,82 +272,22 @@ print_table_for_refs (gboolean print_apps, GPtrArray* refs_array, const char *ar
 }
 
 static gboolean
-print_installed_refs (gboolean app, gboolean runtime, gboolean print_system, gboolean print_user, char **installations, const char *arch, GCancellable *cancellable, GError **error)
+print_installed_refs (gboolean app, gboolean runtime, GPtrArray *dirs, const char *arch, GCancellable *cancellable, GError **error)
 {
   g_autoptr(GPtrArray) refs_array = NULL;
-  gboolean print_all = !print_user && !print_system && (installations == NULL);
+  int i;
 
   refs_array = g_ptr_array_new_with_free_func ((GDestroyNotify) refs_data_free);
-  if (print_user || print_all)
-    {
-      g_autoptr(FlatpakDir) user_dir = NULL;
-      g_auto(GStrv) user_app = NULL;
-      g_auto(GStrv) user_runtime = NULL;
 
-      user_dir =flatpak_dir_get_user ();
-      if (!find_refs_for_dir (user_dir, app ? &user_app : NULL, runtime ? &user_runtime : NULL, cancellable, error))
+  for (i = 0; i < dirs->len; i++)
+    {
+      FlatpakDir *dir = g_ptr_array_index (dirs, i);
+      g_auto(GStrv) apps = NULL;
+      g_auto(GStrv) runtimes = NULL;
+
+      if (!find_refs_for_dir (dir, app ? &apps : NULL, runtime ? &runtimes : NULL, cancellable, error))
         return FALSE;
-      g_ptr_array_add (refs_array, refs_data_new (user_dir, user_app, user_runtime));
-    }
-
-  if (print_all)
-    {
-      g_autoptr(GPtrArray) system_dirs = NULL;
-      int i;
-
-      system_dirs = flatpak_dir_get_system_list (cancellable, error);
-      if (system_dirs == NULL)
-        return FALSE;
-
-      for (i = 0; i < system_dirs->len; i++)
-        {
-          FlatpakDir *dir = g_ptr_array_index (system_dirs, i);
-          g_auto(GStrv) apps = NULL;
-          g_auto(GStrv) runtimes = NULL;
-
-          if (!find_refs_for_dir (dir, app ? &apps : NULL, runtime ? &runtimes : NULL, cancellable, error))
-            return FALSE;
-          g_ptr_array_add (refs_array, refs_data_new (dir, apps, runtimes));
-        }
-    }
-  else
-    {
-      if (print_system)
-        {
-          g_autoptr(FlatpakDir) system_dir = NULL;
-          g_auto(GStrv) system_app = NULL;
-          g_auto(GStrv) system_runtime = NULL;
-
-          system_dir = flatpak_dir_get_system_default ();
-          if (!find_refs_for_dir (system_dir, app ? &system_app : NULL, runtime ? &system_runtime : NULL, cancellable, error))
-            return FALSE;
-          g_ptr_array_add (refs_array, refs_data_new (system_dir, system_app, system_runtime));
-        }
-
-      if (installations != NULL)
-        {
-          g_auto(GStrv) installation_apps = NULL;
-          g_auto(GStrv) installation_runtimes = NULL;
-          int i = 0;
-
-          for (i = 0; installations[i] != NULL; i++)
-            {
-              g_autoptr(FlatpakDir) system_dir = NULL;
-
-              /* Already included the default system installation. */
-              if (print_system && g_strcmp0 (installations[i], "default") == 0)
-                continue;
-
-              system_dir = flatpak_dir_get_system_by_id (installations[i], cancellable, error);
-              if (system_dir == NULL)
-                return FALSE;
-
-              if (!find_refs_for_dir (system_dir, app ? &installation_apps : NULL, runtime ? &installation_runtimes : NULL, cancellable, error))
-                return FALSE;
-
-              g_ptr_array_add (refs_array, refs_data_new (system_dir, installation_apps, installation_runtimes));
-            }
-        }
+      g_ptr_array_add (refs_array, refs_data_new (dir, apps, runtimes));
     }
 
   if (!print_table_for_refs (app, refs_array, arch, cancellable, error))
@@ -343,23 +300,27 @@ gboolean
 flatpak_builtin_list (int argc, char **argv, GCancellable *cancellable, GError **error)
 {
   g_autoptr(GOptionContext) context = NULL;
+  g_autoptr(GPtrArray) dirs = NULL;
 
   context = g_option_context_new (_(" - List installed apps and/or runtimes"));
   g_option_context_set_translation_domain (context, GETTEXT_PACKAGE);
 
-  if (!flatpak_option_context_parse (context, options, &argc, &argv, FLATPAK_BUILTIN_FLAG_NO_DIR, NULL, cancellable, error))
+  if (!flatpak_option_context_parse (context, options, &argc, &argv,
+                                     FLATPAK_BUILTIN_FLAG_ALL_DIRS,
+                                     &dirs, cancellable, error))
     return FALSE;
 
   if (argc > 1)
     return usage_error (context, _("Too many arguments"), error);
 
   if (!opt_app && !opt_runtime)
-    opt_app = TRUE;
+    {
+      opt_app = TRUE;
+      opt_runtime = TRUE;
+    }
 
   if (!print_installed_refs (opt_app, opt_runtime,
-                             opt_system,
-                             opt_user,
-                             opt_installations,
+                             dirs,
                              opt_arch,
                              cancellable, error))
     return FALSE;
@@ -372,5 +333,6 @@ flatpak_complete_list (FlatpakCompletion *completion)
 {
   flatpak_complete_options (completion, global_entries);
   flatpak_complete_options (completion, options);
+  flatpak_complete_options (completion, user_entries);
   return TRUE;
 }
