@@ -43,6 +43,8 @@ static char *opt_command;
 static gboolean opt_devel;
 static gboolean opt_log_session_bus;
 static gboolean opt_log_system_bus;
+static gboolean opt_log_a11y_bus;
+static gboolean opt_file_forwarding;
 static char *opt_runtime;
 static char *opt_runtime_version;
 
@@ -55,6 +57,8 @@ static GOptionEntry options[] = {
   { "runtime-version", 0, 0, G_OPTION_ARG_STRING, &opt_runtime_version, N_("Runtime version to use"), N_("VERSION") },
   { "log-session-bus", 0, 0, G_OPTION_ARG_NONE, &opt_log_session_bus, N_("Log session bus calls"), NULL },
   { "log-system-bus", 0, 0, G_OPTION_ARG_NONE, &opt_log_system_bus, N_("Log system bus calls"), NULL },
+  { "log-a11y-bus", 0, 0, G_OPTION_ARG_NONE, &opt_log_a11y_bus, N_("Log accessibility bus calls"), NULL },
+  { "file-forwarding", 0, 0, G_OPTION_ARG_NONE, &opt_file_forwarding, N_("Enable file forwarding"), NULL },
   { NULL }
 };
 
@@ -109,26 +113,7 @@ flatpak_builtin_run (int argc, char **argv, GCancellable *cancellable, GError **
 
   if (branch == NULL || arch == NULL)
     {
-      g_autofree char *current_ref = NULL;
-      g_autoptr(FlatpakDir) user_dir = flatpak_dir_get_user ();
-
-      current_ref = flatpak_dir_current_ref (user_dir, id, cancellable);
-      if (current_ref == NULL)
-        {
-          g_autoptr(GPtrArray) system_dirs = NULL;
-
-          system_dirs = flatpak_dir_get_system_list (cancellable, error);
-          if (system_dirs == NULL)
-            return FALSE;
-
-          for (i = 0; i < system_dirs->len; i++)
-            {
-              FlatpakDir *dir = g_ptr_array_index (system_dirs, i);
-              current_ref = flatpak_dir_current_ref (dir, id, cancellable);
-              if (current_ref != NULL)
-                break;
-            }
-        }
+      g_autofree char *current_ref = flatpak_find_current_ref (id, NULL, NULL);
 
       if (current_ref)
         {
@@ -165,16 +150,20 @@ flatpak_builtin_run (int argc, char **argv, GCancellable *cancellable, GError **
   if (app_deploy == NULL)
     {
       g_autoptr(FlatpakDeploy) runtime_deploy = NULL;
+      g_autoptr(GError) local_error2 = NULL;
 
       runtime_ref = flatpak_compose_ref (FALSE, id, branch, arch, error);
       if (runtime_ref == NULL)
         return FALSE;
 
-      runtime_deploy = flatpak_find_deploy_for_ref (runtime_ref, cancellable, NULL);
+      runtime_deploy = flatpak_find_deploy_for_ref (runtime_ref, cancellable, &local_error2);
       if (runtime_deploy == NULL)
         {
           /* Report old app-kind error, as its more likely right */
-          g_propagate_error (error, g_steal_pointer (&local_error));
+          if (local_error != NULL)
+            g_propagate_error (error, g_steal_pointer (&local_error));
+          else
+            g_propagate_error (error, g_steal_pointer (&local_error2));
           return FALSE;
         }
       /* Clear app-kind error */
@@ -188,7 +177,9 @@ flatpak_builtin_run (int argc, char **argv, GCancellable *cancellable, GError **
                         opt_runtime_version,
                         (opt_devel ? FLATPAK_RUN_FLAG_DEVEL : 0) |
                         (opt_log_session_bus ? FLATPAK_RUN_FLAG_LOG_SESSION_BUS : 0) |
-                        (opt_log_system_bus ? FLATPAK_RUN_FLAG_LOG_SYSTEM_BUS : 0),
+                        (opt_log_system_bus ? FLATPAK_RUN_FLAG_LOG_SYSTEM_BUS : 0) |
+                        (opt_log_a11y_bus ? FLATPAK_RUN_FLAG_LOG_A11Y_BUS : 0) |
+                        (opt_file_forwarding ? FLATPAK_RUN_FLAG_FILE_FORWARDING : 0),
                         opt_command,
                         &argv[rest_argv_start + 1],
                         rest_argc - 1,
