@@ -31,7 +31,10 @@ G_BEGIN_DECLS
 
 typedef enum {
   FLATPAK_BUILTIN_FLAG_NO_DIR = 1 << 0,
-  FLATPAK_BUILTIN_FLAG_NO_REPO = 1 << 1,
+  FLATPAK_BUILTIN_FLAG_OPTIONAL_REPO = 1 << 1,
+  FLATPAK_BUILTIN_FLAG_ONE_DIR = 1 << 2,
+  FLATPAK_BUILTIN_FLAG_STANDARD_DIRS = 1 << 3,
+  FLATPAK_BUILTIN_FLAG_ALL_DIRS = 1 << 4,
 } FlatpakBuiltinFlags;
 
 gboolean flatpak_option_context_parse (GOptionContext     *context,
@@ -39,7 +42,7 @@ gboolean flatpak_option_context_parse (GOptionContext     *context,
                                        int                *argc,
                                        char             ***argv,
                                        FlatpakBuiltinFlags flags,
-                                       FlatpakDir        **out_dir,
+                                       GPtrArray         **out_dirs,
                                        GCancellable       *cancellable,
                                        GError            **error);
 
@@ -59,6 +62,7 @@ BUILTINPROTO (add_remote)
 BUILTINPROTO (modify_remote)
 BUILTINPROTO (delete_remote)
 BUILTINPROTO (ls_remote)
+BUILTINPROTO (info_remote)
 BUILTINPROTO (list_remotes)
 BUILTINPROTO (install)
 BUILTINPROTO (update)
@@ -83,6 +87,9 @@ BUILTINPROTO (document_unexport)
 BUILTINPROTO (document_info)
 BUILTINPROTO (document_list)
 BUILTINPROTO (override)
+BUILTINPROTO (repo)
+BUILTINPROTO (config)
+BUILTINPROTO (search)
 
 #undef BUILTINPROTO
 
