@@ -33,14 +33,17 @@
 #include "flatpak-builtins.h"
 #include "flatpak-utils.h"
 
-static gboolean opt_verbose;
+static int opt_verbose;
 static gboolean opt_ostree_verbose;
 static gboolean opt_version;
 static gboolean opt_default_arch;
 static gboolean opt_supported_arches;
 static gboolean opt_gl_drivers;
 static gboolean opt_user;
-static char *opt_installation;
+static gboolean opt_system;
+static char **opt_installations;
+
+static gboolean is_in_complete;
 
 typedef struct
 {
@@ -62,6 +65,11 @@ static FlatpakCommand commands[] = {
   { "uninstall", N_("Uninstall an installed application or runtime"), flatpak_builtin_uninstall, flatpak_complete_uninstall },
   { "list", N_("List installed apps and/or runtimes"), flatpak_builtin_list, flatpak_complete_list },
   { "info", N_("Show info for installed app or runtime"), flatpak_builtin_info, flatpak_complete_info },
+  { "config", N_("Configure flatpak"), flatpak_builtin_config, flatpak_complete_config },
+
+   /* translators: please keep the leading newline and space */
+  { N_("\n Finding applications and runtimes") },
+  { "search", N_("Search for remote apps/runtimes"), flatpak_builtin_search, flatpak_complete_search },
 
    /* translators: please keep the leading newline and space */
   { N_("\n Running applications") },
@@ -79,11 +87,13 @@ static FlatpakCommand commands[] = {
 
    /* translators: please keep the leading newline and space */
   { N_("\n Manage remote repositories") },
+  { "remotes", N_("List all configured remotes"), flatpak_builtin_list_remotes, flatpak_complete_list_remotes },
   { "remote-add", N_("Add a new remote repository (by URL)"), flatpak_builtin_add_remote, flatpak_complete_add_remote },
   { "remote-modify", N_("Modify properties of a configured remote"), flatpak_builtin_modify_remote, flatpak_complete_modify_remote },
   { "remote-delete", N_("Delete a configured remote"), flatpak_builtin_delete_remote, flatpak_complete_delete_remote },
-  { "remote-list", N_("List all configured remotes"), flatpak_builtin_list_remotes, flatpak_complete_list_remotes },
+  { "remote-list", NULL, flatpak_builtin_list_remotes, flatpak_complete_list_remotes, TRUE },
   { "remote-ls", N_("List contents of a configured remote"), flatpak_builtin_ls_remote, flatpak_complete_ls_remote },
+  { "remote-info", N_("Show information about a remote app or runtime"), flatpak_builtin_info_remote, flatpak_complete_info_remote },
 
    /* translators: please keep the leading newline and space */
   { N_("\n Build applications") },
@@ -91,17 +101,29 @@ static FlatpakCommand commands[] = {
   { "build", N_("Run a build command inside the build dir"), flatpak_builtin_build, flatpak_complete_build  },
   { "build-finish", N_("Finish a build dir for export"), flatpak_builtin_build_finish, flatpak_complete_build_finish },
   { "build-export", N_("Export a build dir to a repository"), flatpak_builtin_build_export, flatpak_complete_build_export },
-  { "build-bundle", N_("Create a bundle file from a build directory"), flatpak_builtin_build_bundle, flatpak_complete_build_bundle },
+  { "build-bundle", N_("Create a bundle file from a ref in a local repository"), flatpak_builtin_build_bundle, flatpak_complete_build_bundle },
   { "build-import-bundle", N_("Import a bundle file"), flatpak_builtin_build_import, flatpak_complete_build_import },
   { "build-sign", N_("Sign an application or runtime"), flatpak_builtin_build_sign, flatpak_complete_build_sign },
   { "build-update-repo", N_("Update the summary file in a repository"), flatpak_builtin_build_update_repo, flatpak_complete_build_update_repo },
   { "build-commit-from", N_("Create new commit based on existing ref"), flatpak_builtin_build_commit_from, flatpak_complete_build_commit_from },
+  { "repo", N_("Print information about a repo"), flatpak_builtin_repo, flatpak_complete_repo },
 
   { NULL }
 };
 
+static gboolean
+opt_verbose_cb (const gchar *option_name,
+                const gchar *value,
+                gpointer     data,
+                GError     **error)
+{
+  opt_verbose++;
+  return TRUE;
+}
+
+
 GOptionEntry global_entries[] = {
-  { "verbose", 'v', 0, G_OPTION_ARG_NONE, &opt_verbose, N_("Print debug information during command processing"), NULL },
+  { "verbose", 'v', G_OPTION_FLAG_NO_ARG, G_OPTION_ARG_CALLBACK, &opt_verbose_cb, N_("Print debug information during command processing, -vv for more detail"), NULL },
   { "ostree-verbose", 0, 0, G_OPTION_ARG_NONE, &opt_ostree_verbose, N_("Print OSTree debug information during command processing"), NULL },
   { "help", '?', G_OPTION_FLAG_HIDDEN, G_OPTION_ARG_NONE, N_("Show help options"), NULL, NULL },
   { NULL }
@@ -117,8 +139,8 @@ static GOptionEntry empty_entries[] = {
 
 GOptionEntry user_entries[] = {
   { "user", 0, 0, G_OPTION_ARG_NONE, &opt_user, N_("Work on user installations"), NULL },
-  { "system", 0, G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &opt_user, N_("Work on system-wide installations (default)"), NULL },
-  { "installation", 0, 0, G_OPTION_ARG_STRING, &opt_installation, N_("Work on a specific system-wide installation"), NULL },
+  { "system", 0, 0, G_OPTION_ARG_NONE, &opt_system, N_("Work on system-wide installations (default)"), NULL },
+  { "installation", 0, 0, G_OPTION_ARG_STRING_ARRAY, &opt_installations, N_("Work on specific system-wide installation(s)"), N_("NAME") },
   { NULL }
 };
 
@@ -130,12 +152,12 @@ message_handler (const gchar   *log_domain,
 {
   /* Make this look like normal console output */
   if (log_level & G_LOG_LEVEL_DEBUG)
-    g_printerr ("XA: %s\n", message);
+    g_printerr ("F: %s\n", message);
   else
     g_printerr ("%s: %s\n", g_get_prgname (), message);
 }
 
-GOptionContext *
+static GOptionContext *
 flatpak_option_context_new_with_commands (FlatpakCommand *commands)
 {
   GOptionContext *context;
@@ -171,7 +193,7 @@ flatpak_option_context_new_with_commands (FlatpakCommand *commands)
   return context;
 }
 
-int
+static int
 flatpak_usage (FlatpakCommand *commands,
                gboolean        is_error)
 {
@@ -200,11 +222,32 @@ flatpak_option_context_parse (GOptionContext     *context,
                               int                *argc,
                               char             ***argv,
                               FlatpakBuiltinFlags flags,
-                              FlatpakDir        **out_dir,
+                              GPtrArray         **out_dirs,
                               GCancellable       *cancellable,
                               GError            **error)
 {
-  g_autoptr(FlatpakDir) dir = NULL;
+  g_autoptr(GPtrArray) dirs = NULL;
+
+  if (!(flags & FLATPAK_BUILTIN_FLAG_NO_DIR) &&
+      !(flags & FLATPAK_BUILTIN_FLAG_ONE_DIR) &&
+      !(flags & FLATPAK_BUILTIN_FLAG_STANDARD_DIRS) &&
+      !(flags & FLATPAK_BUILTIN_FLAG_ALL_DIRS))
+    g_assert_not_reached ();
+
+  if (flags & FLATPAK_BUILTIN_FLAG_NO_DIR &&
+      (flags & FLATPAK_BUILTIN_FLAG_ONE_DIR ||
+       flags & FLATPAK_BUILTIN_FLAG_STANDARD_DIRS ||
+       flags & FLATPAK_BUILTIN_FLAG_ALL_DIRS))
+    g_assert_not_reached ();
+
+  if (flags & FLATPAK_BUILTIN_FLAG_ONE_DIR &&
+      (flags & FLATPAK_BUILTIN_FLAG_STANDARD_DIRS ||
+       flags & FLATPAK_BUILTIN_FLAG_ALL_DIRS))
+    g_assert_not_reached ();
+
+  if (flags & FLATPAK_BUILTIN_FLAG_STANDARD_DIRS &&
+      flags & FLATPAK_BUILTIN_FLAG_ALL_DIRS)
+    g_assert_not_reached ();
 
   if (!(flags & FLATPAK_BUILTIN_FLAG_NO_DIR))
     g_option_context_add_main_entries (context, user_entries, NULL);
@@ -217,6 +260,18 @@ flatpak_option_context_parse (GOptionContext     *context,
   if (!g_option_context_parse (context, argc, argv, error))
     return FALSE;
 
+  /* We never want verbose output in the complete case, that breaks completion */
+  if (!is_in_complete)
+    {
+      if (opt_verbose > 0)
+        g_log_set_handler (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, message_handler, NULL);
+      if (opt_verbose > 1)
+        g_log_set_handler (G_LOG_DOMAIN "2", G_LOG_LEVEL_DEBUG, message_handler, NULL);
+
+      if (opt_ostree_verbose)
+        g_log_set_handler ("OSTree", G_LOG_LEVEL_DEBUG, message_handler, NULL);
+    }
+
   if (opt_version)
     {
       g_print ("%s\n", PACKAGE_STRING);
@@ -249,33 +304,102 @@ flatpak_option_context_parse (GOptionContext     *context,
 
   if (!(flags & FLATPAK_BUILTIN_FLAG_NO_DIR))
     {
-      if (opt_user)
-        dir = flatpak_dir_get_user ();
-      else if (opt_installation == NULL)
-        dir = flatpak_dir_get_system_default ();
-      else
+      dirs = g_ptr_array_new_with_free_func ((GDestroyNotify) g_object_unref);
+      int i;
+
+      if (!(flags & FLATPAK_BUILTIN_FLAG_ONE_DIR))
         {
-          dir = flatpak_dir_get_system_by_id (opt_installation, cancellable, error);
-          if (dir == NULL)
-            return FALSE;
+          /*
+           * FLATPAK_BUILTIN_FLAG_STANDARD_DIRS or FLATPAK_BUILTIN_FLAG_ALL_DIRS
+           * must be set.
+           */
+          if (opt_user || (!opt_system && opt_installations == NULL))
+            g_ptr_array_add (dirs, flatpak_dir_get_user ());
+
+          if (opt_system || (!opt_user && opt_installations == NULL))
+            g_ptr_array_add (dirs, flatpak_dir_get_system_default ());
+
+          if (opt_installations != NULL)
+            {
+              for (i = 0; opt_installations[i] != NULL; i++)
+                {
+                  FlatpakDir *installation_dir = NULL;
+
+                  /* Already included the default system installation */
+                  if (opt_system && g_strcmp0 (opt_installations[i], "default") == 0)
+                    continue;
+
+                  installation_dir = flatpak_dir_get_system_by_id (opt_installations[i], cancellable, error);
+                  if (installation_dir == NULL)
+                    return FALSE;
+
+                  g_ptr_array_add (dirs, installation_dir);
+                }
+            }
+
+          if (flags & FLATPAK_BUILTIN_FLAG_ALL_DIRS &&
+              opt_installations == NULL && !opt_user && !opt_system)
+            {
+              g_autoptr(GPtrArray) system_dirs = NULL;
+
+              g_ptr_array_set_size (dirs, 0);
+              g_ptr_array_add (dirs, flatpak_dir_get_user ());
+
+              system_dirs = flatpak_dir_get_system_list (cancellable, error);
+              if (system_dirs == NULL)
+                return FALSE;
+
+              for (i = 0; i < system_dirs->len; i++)
+                {
+                  FlatpakDir *dir = g_ptr_array_index (system_dirs, i);
+                  g_ptr_array_add (dirs, g_object_ref (dir));
+                }
+            }
+        }
+      else /* FLATPAK_BUILTIN_FLAG_ONE_DIR */
+        {
+          FlatpakDir *dir;
+
+          if (opt_system || (!opt_user && opt_installations == NULL))
+            dir = flatpak_dir_get_system_default ();
+          else if (opt_user)
+            dir = flatpak_dir_get_user ();
+          else if (opt_installations != NULL)
+            {
+              if (g_strv_length (opt_installations) > 1)
+                return usage_error (context, _("The --installation option was used multiple times "
+                                               "for a command that works on one installation"), error);
+              dir = flatpak_dir_get_system_by_id (opt_installations[0], cancellable, error);
+              if (dir == NULL)
+                return FALSE;
+            }
+          else
+            g_assert_not_reached ();
+
+          g_ptr_array_add (dirs, dir);
         }
 
-      if (!flatpak_dir_ensure_path (dir, cancellable, error))
-        return FALSE;
+      for (i = 0; i < dirs->len; i++)
+        {
+          FlatpakDir *dir = g_ptr_array_index (dirs, i);
 
-      if (!(flags & FLATPAK_BUILTIN_FLAG_NO_REPO) &&
-          !flatpak_dir_ensure_repo (dir, cancellable, error))
-        return FALSE;
+          if (flags & FLATPAK_BUILTIN_FLAG_OPTIONAL_REPO)
+            {
+              if (!flatpak_dir_maybe_ensure_repo (dir, cancellable, error))
+                return FALSE;
+            }
+          else
+            {
+              if (!flatpak_dir_ensure_repo (dir, cancellable, error))
+                return FALSE;
+            }
+
+          flatpak_log_dir_access (dir);
+        }
     }
 
-  if (opt_verbose)
-    g_log_set_handler (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, message_handler, NULL);
-
-  if (opt_ostree_verbose)
-    g_log_set_handler ("OSTree", G_LOG_LEVEL_DEBUG, message_handler, NULL);
-
-  if (out_dir)
-    *out_dir = g_steal_pointer (&dir);
+  if (out_dirs)
+    *out_dirs = g_steal_pointer (&dirs);
 
   return TRUE;
 }
@@ -290,7 +414,7 @@ usage_error (GOptionContext *context, const char *message, GError **error)
   return FALSE;
 }
 
-FlatpakCommand *
+static FlatpakCommand *
 extract_command (int     *argc,
                  char   **argv,
                  const char **command_name_out)
@@ -338,7 +462,7 @@ extract_command (int     *argc,
 }
 
 
-int
+static int
 flatpak_run (int      argc,
              char   **argv,
              GError **res_error)
@@ -408,6 +532,8 @@ complete (int    argc,
   FlatpakCompletion *completion;
   const char *command_name = NULL;
 
+  is_in_complete = TRUE;
+
   completion = flatpak_completion_new (argv[2], argv[3], argv[4]);
   if (completion == NULL)
     return 1;
@@ -479,13 +605,12 @@ main (int    argc,
 
   if (error != NULL)
     {
-      int is_tty = isatty (1);
       const char *prefix = "";
       const char *suffix = "";
-      if (is_tty)
+      if (flatpak_fancy_output ())
         {
-          prefix = "\x1b[31m\x1b[1m"; /* red, bold */
-          suffix = "\x1b[22m\x1b[0m"; /* bold off, color reset */
+          prefix = FLATPAK_ANSI_RED FLATPAK_ANSI_BOLD_ON;
+          suffix = FLATPAK_ANSI_BOLD_OFF FLATPAK_ANSI_COLOR_RESET;
         }
       g_printerr ("%s%s %s%s\n", prefix, _("error:"), suffix, error->message);
       g_error_free (error);
