@@ -20,12 +20,12 @@
 
 #include "config.h"
 
+#include <stdio.h>
 #include <glib/gi18n.h>
 
 #include "flatpak-transaction.h"
 #include "flatpak-utils.h"
 #include "flatpak-builtins-utils.h"
-#include "flatpak-oci-registry.h"
 #include "flatpak-error.h"
 
 typedef struct FlatpakTransactionOp FlatpakTransactionOp;
@@ -40,6 +40,7 @@ typedef enum {
 struct FlatpakTransactionOp {
   char *remote;
   char *ref;
+  /* NULL means unspecified (normally keep whatever was there before), [] means force everything */
   char **subpaths;
   char *commit;
   GFile *bundle;
@@ -52,14 +53,23 @@ struct FlatpakTransaction {
   GHashTable *refs;
   GPtrArray *system_dirs;
   GList *ops;
+  GPtrArray *added_origin_remotes;
 
   gboolean no_interaction;
   gboolean no_pull;
   gboolean no_deploy;
+  gboolean no_static_deltas;
   gboolean add_deps;
   gboolean add_related;
+  gboolean reinstall;
 };
 
+static gboolean
+remote_name_is_file (const char *remote_name)
+{
+  return remote_name != NULL &&
+    g_str_has_prefix (remote_name, "file://");
+}
 
 /* Check if the ref is in the dir, or in the system dir, in case its a
  * user-dir or another system-wide installation. We want to avoid depending
@@ -155,24 +165,35 @@ flatpak_transaction_operation_free (FlatpakTransactionOp *self)
   g_free (self);
 }
 
+gboolean
+flatpak_transaction_is_empty (FlatpakTransaction  *self)
+{
+  return self->ops == NULL;
+}
+
 FlatpakTransaction *
 flatpak_transaction_new (FlatpakDir *dir,
                          gboolean no_interaction,
                          gboolean no_pull,
                          gboolean no_deploy,
+                         gboolean no_static_deltas,
                          gboolean add_deps,
-                         gboolean add_related)
+                         gboolean add_related,
+                         gboolean reinstall)
 {
   FlatpakTransaction *t = g_new0 (FlatpakTransaction, 1);
 
   t->dir = g_object_ref (dir);
   t->refs = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+  t->added_origin_remotes = g_ptr_array_new_with_free_func (g_free);
 
   t->no_interaction = no_interaction;
   t->no_pull = no_pull;
   t->no_deploy = no_deploy;
+  t->no_static_deltas = no_static_deltas;
   t->add_deps = add_deps;
   t->add_related = add_related;
+  t->reinstall = reinstall;
   return t;
 }
 
@@ -183,13 +204,15 @@ flatpak_transaction_free (FlatpakTransaction *self)
   g_list_free_full (self->ops, (GDestroyNotify)flatpak_transaction_operation_free);
   g_object_unref (self->dir);
 
+  g_ptr_array_unref (self->added_origin_remotes);
+
   if (self->system_dirs != NULL)
     g_ptr_array_free (self->system_dirs, TRUE);
 
   g_free (self);
 }
 
-gboolean
+static gboolean
 flatpak_transaction_contains_ref (FlatpakTransaction *self,
                                   const char *ref)
 {
@@ -242,7 +265,7 @@ kind_to_str (FlatpakTransactionOpKind kind)
   return "unknown";
 }
 
-FlatpakTransactionOp *
+static FlatpakTransactionOp *
 flatpak_transaction_add_op (FlatpakTransaction *self,
                             const char *remote,
                             const char *ref,
@@ -264,13 +287,8 @@ flatpak_transaction_add_op (FlatpakTransaction *self,
   op = g_hash_table_lookup (self->refs, ref);
   if (op != NULL)
     {
-      /* Only override subpaths if already specified,
-         we always want the un-subpathed to win if specified. */
-      if (op->subpaths != NULL && op->subpaths[0] != NULL && subpaths != NULL)
-        {
-          g_strfreev (op->subpaths);
-          op->subpaths = g_strdupv ((char **)subpaths);
-        }
+      g_auto(GStrv) old_subpaths = op->subpaths;
+      op->subpaths = flatpak_subpaths_merge (old_subpaths, (char **)subpaths);
 
       return op;
     }
@@ -292,7 +310,7 @@ ask_for_remote (FlatpakTransaction *self, const char **remotes)
   if (self->no_interaction)
     {
       chosen = 1;
-      g_print ("Found in remote %s\n", remotes[0]);
+      g_print (_("Found in remote %s\n"), remotes[0]);
     }
   else if (n_remotes == 1)
     {
@@ -455,6 +473,28 @@ flatpak_transaction_add_ref (FlatpakTransaction *self,
   g_autofree char *remote_metadata = NULL;
   g_autoptr(GKeyFile) metakey = NULL;
   g_autoptr(GError) local_error = NULL;
+  g_autofree char *origin_remote = NULL;
+
+  if (remote_name_is_file (remote))
+    {
+      g_auto(GStrv) parts = NULL;
+      parts = g_strsplit (ref, "/", -1);
+
+      origin_remote = flatpak_dir_create_origin_remote (self->dir,
+                                                        remote, /* uri */
+                                                        parts[1],
+                                                        "Local repo",
+                                                        ref,
+                                                        NULL,
+                                                        NULL,
+                                                        NULL, error);
+      if (origin_remote == NULL)
+        return FALSE;
+
+      g_ptr_array_add (self->added_origin_remotes, g_strdup (origin_remote));
+
+      remote = origin_remote;
+    }
 
   pref = strchr (ref, '/') + 1;
 
@@ -477,11 +517,20 @@ flatpak_transaction_add_ref (FlatpakTransaction *self,
   else if (kind == FLATPAK_TRANSACTION_OP_KIND_INSTALL)
     {
       g_assert (remote != NULL);
-      if (dir_ref_is_installed (self->dir, ref, NULL, NULL))
+      if (!self->reinstall &&
+          dir_ref_is_installed (self->dir, ref, &origin, NULL))
         {
-          g_set_error (error, FLATPAK_ERROR, FLATPAK_ERROR_ALREADY_INSTALLED,
-                       _("%s already installed"), pref);
-          return FALSE;
+          if (strcmp (remote, origin) == 0)
+            {
+              g_printerr (_("%s already installed, skipping\n"), pref);
+              return TRUE;
+            }
+          else
+            {
+              g_set_error (error, FLATPAK_ERROR, FLATPAK_ERROR_ALREADY_INSTALLED,
+                           _("%s is already installed from other remote (%s)"), pref, origin);
+              return FALSE;
+            }
         }
     }
 
@@ -491,7 +540,7 @@ flatpak_transaction_add_ref (FlatpakTransaction *self,
         metadata = remote_metadata;
       else
         {
-          g_print ("Warning: Can't find dependencies: %s\n", local_error->message);
+          g_print (_("Warning: Can't find dependencies: %s\n"), local_error->message);
           g_clear_error (&local_error);
         }
     }
@@ -518,7 +567,7 @@ flatpak_transaction_add_ref (FlatpakTransaction *self,
       if (required_version)
         {
           if (sscanf (required_version, "%d.%d.%d", &required_major, &required_minor, &required_micro) != 3)
-            g_print ("Invalid require-flatpak argument %s\n", required_version);
+            g_print (_("Invalid require-flatpak argument %s\n"), required_version);
           else
             {
               if (required_major > PACKAGE_MAJOR_VERSION ||
@@ -582,66 +631,6 @@ flatpak_transaction_add_install_bundle (FlatpakTransaction *self,
   return flatpak_transaction_add_ref (self, remote, ref, NULL, NULL, FLATPAK_TRANSACTION_OP_KIND_BUNDLE, file, metadata, error);
 }
 
-gboolean
-flatpak_transaction_add_install_oci (FlatpakTransaction  *self,
-                                     const char         *uri,
-                                     const char         *tag,
-                                     GError             **error)
-{
-  GHashTable *annotations;
-  g_autofree char *ref = NULL;
-  g_autofree char *checksum = NULL;
-  g_autoptr(FlatpakOciManifest) manifest = NULL;
-  g_autoptr(FlatpakOciRegistry) registry = NULL;
-  const char *all_paths[] = { NULL };
-  g_autofree char *remote = NULL;
-  g_autofree char *title = NULL;
-  g_auto(GStrv) parts = NULL;
-  g_autofree char *id = NULL;
-
-  registry = flatpak_oci_registry_new (uri, FALSE, -1, NULL, error);
-  if (registry == NULL)
-    return FALSE;
-
-  manifest = flatpak_oci_registry_chose_image (registry, tag, NULL,
-                                               NULL, error);
-  if (manifest == NULL)
-    return FALSE;
-
-  /* TODO: Extract runtime dependencies and related refs */
-  annotations = flatpak_oci_manifest_get_annotations (manifest);
-  if (annotations)
-    flatpak_oci_parse_commit_annotations (annotations, NULL,
-                                          NULL, NULL,
-                                          &ref, &checksum, NULL,
-                                          NULL);
-
-  if (ref == NULL)
-    return flatpak_fail (error, _("OCI image is not a flatpak (missing ref)"));
-
-  parts = flatpak_decompose_ref (ref, error);
-  if (parts == NULL)
-    return FALSE;
-
-  title = g_strdup_printf ("OCI remote for %s", parts[1]);
-
-  id = g_strdup_printf ("oci-%s", parts[1]);
-
-  remote = flatpak_dir_create_origin_remote (self->dir, NULL,
-                                             id, title,
-                                             ref, uri, tag, NULL,
-                                             NULL, error);
-  if (remote == NULL)
-    return FALSE;
-
-  if (!flatpak_dir_recreate_repo (self->dir, NULL, error))
-    return FALSE;
-
-  g_debug ("Added OCI origin remote %s", remote);
-
-  return flatpak_transaction_add_ref (self, remote, ref, all_paths, checksum, FLATPAK_TRANSACTION_OP_KIND_INSTALL, NULL, NULL, error);
-}
-
 gboolean
 flatpak_transaction_add_update (FlatpakTransaction *self,
                                 const char *ref,
@@ -649,9 +638,63 @@ flatpak_transaction_add_update (FlatpakTransaction *self,
                                 const char *commit,
                                 GError **error)
 {
+  const char *all_paths[] = { NULL };
+
+  /* If specify an empty subpath, that means all subpaths */
+  if (subpaths != NULL && subpaths[0] != NULL && subpaths[0][0] == 0)
+    subpaths = all_paths;
+
   return flatpak_transaction_add_ref (self, NULL, ref, subpaths, commit, FLATPAK_TRANSACTION_OP_KIND_UPDATE, NULL, NULL, error);
 }
 
+gboolean
+flatpak_transaction_update_metadata (FlatpakTransaction  *self,
+                                     gboolean             all_remotes,
+                                     GCancellable        *cancellable,
+                                     GError             **error)
+{
+  g_auto(GStrv) remotes = NULL;
+  int i;
+  GList *l;
+
+  /* Collect all dir+remotes used in this transaction */
+
+  if (all_remotes)
+    {
+      remotes = flatpak_dir_list_remotes (self->dir, NULL, error);
+      if (remotes == NULL)
+        return FALSE;
+    }
+  else
+    {
+      g_autoptr(GHashTable) ht = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+      for (l = self->ops; l != NULL; l = l->next)
+        {
+          FlatpakTransactionOp *op = l->data;
+          g_hash_table_add (ht, g_strdup (op->remote));
+        }
+      remotes = (char **)g_hash_table_get_keys_as_array (ht, NULL);
+      g_hash_table_steal_all (ht); /* Move ownership to remotes */
+    }
+
+  /* Update metadata for said remotes */
+  for (i = 0; remotes[i] != NULL; i++)
+    {
+      char *remote = remotes[i];
+      g_autoptr(GError) my_error = NULL;
+
+      g_debug ("Updating remote metadata for %s", remote);
+      if (!flatpak_dir_update_remote_configuration (self->dir, remote, cancellable, &my_error))
+        g_printerr (_("Error updating remote metadata for '%s': %s\n"), remote, my_error->message);
+    }
+
+  /* Reload changed configuration */
+  if (!flatpak_dir_recreate_repo (self->dir, cancellable, error))
+    return FALSE;
+
+  return TRUE;
+}
+
 gboolean
 flatpak_transaction_run (FlatpakTransaction *self,
                          gboolean stop_on_first_error,
@@ -660,6 +703,7 @@ flatpak_transaction_run (FlatpakTransaction *self,
 {
   GList *l;
   gboolean succeeded = TRUE;
+  int i;
 
   self->ops = g_list_reverse (self->ops);
 
@@ -667,10 +711,11 @@ flatpak_transaction_run (FlatpakTransaction *self,
     {
       FlatpakTransactionOp *op = l->data;
       g_autoptr(GError) local_error = NULL;
-      gboolean res;
+      gboolean res = TRUE;
       const char *pref;
       const char *opname;
       FlatpakTransactionOpKind kind;
+      FlatpakTerminalProgress terminal_progress = { 0 };
 
       kind = op->kind;
       if (kind == FLATPAK_TRANSACTION_OP_KIND_INSTALL_OR_UPDATE)
@@ -679,14 +724,6 @@ flatpak_transaction_run (FlatpakTransaction *self,
 
           if (dir_ref_is_installed (self->dir, op->ref, NULL, &deploy_data))
             {
-              g_autofree const char **current_subpaths = NULL;
-
-              /* When we update a dependency, we always inherit the subpaths
-                 rather than use the default. */
-              g_strfreev (op->subpaths);
-              current_subpaths = flatpak_deploy_data_get_subpaths (deploy_data);
-              op->subpaths = g_strdupv ((char **)current_subpaths);
-
               /* Don't use the remote from related ref on update, always use
                  the current remote. */
               g_free (op->remote);
@@ -700,52 +737,91 @@ flatpak_transaction_run (FlatpakTransaction *self,
 
       pref = strchr (op->ref, '/') + 1;
 
+
       if (kind == FLATPAK_TRANSACTION_OP_KIND_INSTALL)
         {
+          g_autoptr(OstreeAsyncProgress) progress = flatpak_progress_new (flatpak_terminal_progress_cb, &terminal_progress);
           opname = _("install");
-          g_print (_("Installing: %s from %s\n"), pref, op->remote);
-          res = flatpak_dir_install (self->dir,
+          if (flatpak_dir_is_user (self->dir))
+            g_print (_("Installing for user: %s from %s\n"), pref, op->remote);
+          else
+            g_print (_("Installing: %s from %s\n"), pref, op->remote);
+          res = flatpak_dir_install (self->dir ,
                                      self->no_pull,
                                      self->no_deploy,
+                                     self->no_static_deltas,
+                                     self->reinstall,
                                      op->ref, op->remote,
                                      (const char **)op->subpaths,
-                                     NULL,
+                                     progress,
                                      cancellable, &local_error);
+          ostree_async_progress_finish (progress);
+          flatpak_terminal_progress_end (&terminal_progress);
         }
       else if (kind == FLATPAK_TRANSACTION_OP_KIND_UPDATE)
         {
+          g_auto(OstreeRepoFinderResultv) check_results = NULL;
+
           opname = _("update");
-          g_print (_("Updating: %s from %s\n"), pref, op->remote);
-          res = flatpak_dir_update (self->dir,
-                                    self->no_pull,
-                                    self->no_deploy,
-                                    op->ref, op->remote, op->commit,
-                                    (const char **)op->subpaths,
-                                    NULL,
-                                    cancellable, &local_error);
-
-          if (res)
+          g_autofree char *target_commit = flatpak_dir_check_for_update (self->dir, op->ref, op->remote, op->commit,
+                                                                         (const char **)op->subpaths,
+                                                                         self->no_pull,
+                                                                         &check_results,
+                                                                         cancellable, &local_error);
+          if (target_commit != NULL)
             {
-              g_autoptr(GVariant) deploy_data = NULL;
-              g_autofree char *commit = NULL;
-              deploy_data = flatpak_dir_get_deploy_data (self->dir, op->ref, NULL, NULL);
-              commit = g_strndup (flatpak_deploy_data_get_commit (deploy_data), 12);
-              g_print (_("Now at %s.\n"), commit);
+              if (flatpak_dir_is_user (self->dir))
+                g_print (_("Updating for user: %s from %s\n"), pref, op->remote);
+              else
+                g_print (_("Updating: %s from %s\n"), pref, op->remote);
+              g_autoptr(OstreeAsyncProgress) progress = flatpak_progress_new (flatpak_terminal_progress_cb, &terminal_progress);
+              res = flatpak_dir_update (self->dir,
+                                        self->no_pull,
+                                        self->no_deploy,
+                                        self->no_static_deltas,
+                                        op->commit != NULL, /* Allow downgrade if we specify commit */
+                                        op->ref, op->remote, target_commit,
+                                        (const OstreeRepoFinderResult * const *) check_results,
+                                        (const char **)op->subpaths,
+                                        progress,
+                                        cancellable, &local_error);
+              ostree_async_progress_finish (progress);
+              flatpak_terminal_progress_end (&terminal_progress);
+              if (res)
+                {
+                  g_autoptr(GVariant) deploy_data = NULL;
+                  g_autofree char *commit = NULL;
+                  deploy_data = flatpak_dir_get_deploy_data (self->dir, op->ref, NULL, NULL);
+                  commit = g_strndup (flatpak_deploy_data_get_commit (deploy_data), 12);
+                  g_print (_("Now at %s.\n"), commit);
+                }
+
+              /* Handle noop-updates */
+              if (!res && g_error_matches (local_error, FLATPAK_ERROR, FLATPAK_ERROR_ALREADY_INSTALLED))
+                {
+                  g_print (_("No updates.\n"));
+                  res = TRUE;
+                  g_clear_error (&local_error);
+                }
             }
-
-          /* Handle noop-updates */
-          if (!res && g_error_matches (local_error, FLATPAK_ERROR, FLATPAK_ERROR_ALREADY_INSTALLED))
+          else
             {
-              g_print (_("No updates.\n"));
-              res = TRUE;
-              g_clear_error (&local_error);
+              res = FALSE;
+              if (g_error_matches (local_error, FLATPAK_ERROR, FLATPAK_ERROR_ALREADY_INSTALLED))
+                {
+                  res = TRUE;
+                  g_clear_error (&local_error);
+                }
             }
         }
       else if (kind == FLATPAK_TRANSACTION_OP_KIND_BUNDLE)
         {
           g_autofree char *bundle_basename = g_file_get_basename (op->bundle);
           opname = _("install bundle");
-          g_print (_("Installing: %s from bundle %s\n"), pref, bundle_basename);
+          if (flatpak_dir_is_user (self->dir))
+            g_print (_("Installing for user: %s from bundle %s\n"), pref, bundle_basename);
+          else
+            g_print (_("Installing: %s from bundle %s\n"), pref, bundle_basename);
           res = flatpak_dir_install_bundle (self->dir, op->bundle,
                                             op->remote, NULL,
                                             cancellable, &local_error);
@@ -772,11 +848,17 @@ flatpak_transaction_run (FlatpakTransaction *self,
             }
           else
             {
+              succeeded = FALSE;
               g_propagate_error (error, g_steal_pointer (&local_error));
-              return FALSE;
+              goto out;
             }
         }
     }
 
+ out:
+
+  for (i = 0; i < self->added_origin_remotes->len; i++)
+    flatpak_dir_prune_origin_remote (self->dir, g_ptr_array_index (self->added_origin_remotes, i));
+
   return succeeded;
 }
