@@ -25,6 +25,7 @@
 
 #include "libglnx/libglnx.h"
 #include <flatpak-common-types.h>
+#include <flatpak-context.h>
 
 #define FLATPAK_TYPE_DIR flatpak_dir_get_type ()
 #define FLATPAK_DIR(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), FLATPAK_TYPE_DIR, FlatpakDir))
@@ -41,12 +42,17 @@ GType flatpak_deploy_get_type (void);
 #define FLATPAK_REF_VERSION_KEY "Version"
 #define FLATPAK_REF_URL_KEY "Url"
 #define FLATPAK_REF_RUNTIME_REPO_KEY "RuntimeRepo"
+#define FLATPAK_REF_SUGGEST_REMOTE_NAME_KEY "SuggestRemoteName"
 #define FLATPAK_REF_TITLE_KEY "Title"
 #define FLATPAK_REF_GPGKEY_KEY "GPGKey"
 #define FLATPAK_REF_IS_RUNTIME_KEY "IsRuntime"
 #define FLATPAK_REF_NAME_KEY "Name"
 #define FLATPAK_REF_BRANCH_KEY "Branch"
 
+#ifdef FLATPAK_ENABLE_P2P
+#define FLATPAK_REF_COLLECTION_ID_KEY "CollectionID"
+#endif  /* FLATPAK_ENABLE_P2P */
+
 #define FLATPAK_REPO_GROUP "Flatpak Repo"
 #define FLATPAK_REPO_VERSION_KEY "Version"
 #define FLATPAK_REPO_URL_KEY "Url"
@@ -55,8 +61,16 @@ GType flatpak_deploy_get_type (void);
 #define FLATPAK_REPO_GPGKEY_KEY "GPGKey"
 #define FLATPAK_REPO_NODEPS_KEY "NoDeps"
 
+#ifdef FLATPAK_ENABLE_P2P
+#define FLATPAK_REPO_COLLECTION_ID_KEY "CollectionID"
+#endif  /* FLATPAK_ENABLE_P2P */
+
+#define FLATPAK_DEFAULT_UPDATE_FREQUENCY 100
+#define FLATPAK_CLI_UPDATE_FREQUENCY 300
+
 typedef struct
 {
+  char           *collection_id;  /* (nullable) */
   char           *ref;
   char           *commit;
   char          **subpaths;
@@ -75,9 +89,10 @@ typedef enum {
   FLATPAK_HELPER_DEPLOY_FLAGS_UPDATE = 1 << 0,
   FLATPAK_HELPER_DEPLOY_FLAGS_NO_DEPLOY = 1 << 1,
   FLATPAK_HELPER_DEPLOY_FLAGS_LOCAL_PULL = 1 << 2,
+  FLATPAK_HELPER_DEPLOY_FLAGS_REINSTALL = 1 << 3,
 } FlatpakHelperDeployFlags;
 
-#define FLATPAK_HELPER_DEPLOY_FLAGS_ALL (FLATPAK_HELPER_DEPLOY_FLAGS_UPDATE|FLATPAK_HELPER_DEPLOY_FLAGS_NO_DEPLOY|FLATPAK_HELPER_DEPLOY_FLAGS_LOCAL_PULL)
+#define FLATPAK_HELPER_DEPLOY_FLAGS_ALL (FLATPAK_HELPER_DEPLOY_FLAGS_UPDATE|FLATPAK_HELPER_DEPLOY_FLAGS_NO_DEPLOY|FLATPAK_HELPER_DEPLOY_FLAGS_LOCAL_PULL|FLATPAK_HELPER_DEPLOY_FLAGS_REINSTALL)
 
 typedef enum {
   FLATPAK_HELPER_UNINSTALL_FLAGS_NONE = 0,
@@ -94,11 +109,25 @@ typedef enum {
 
 #define FLATPAK_HELPER_CONFIGURE_REMOTE_FLAGS_ALL (FLATPAK_HELPER_CONFIGURE_REMOTE_FLAGS_FORCE_REMOVE)
 
+typedef enum {
+  FLATPAK_HELPER_CONFIGURE_FLAGS_NONE = 0,
+  FLATPAK_HELPER_CONFIGURE_FLAGS_UNSET = 1 << 0,
+} FlatpakHelperConfigureFlags;
+
+#define FLATPAK_HELPER_CONFIGURE_FLAGS_ALL (FLATPAK_HELPER_CONFIGURE_FLAGS_UNSET)
+
+typedef enum {
+  FLATPAK_HELPER_UPDATE_REMOTE_FLAGS_NONE = 0,
+} FlatpakHelperUpdateRemoteFlags;
+
+#define FLATPAK_HELPER_UPDATE_REMOTE_FLAGS_ALL (0)
+
 typedef enum {
   FLATPAK_PULL_FLAGS_NONE = 0,
   FLATPAK_PULL_FLAGS_DOWNLOAD_EXTRA_DATA = 1 << 0,
   FLATPAK_PULL_FLAGS_SIDELOAD_EXTRA_DATA = 1 << 1,
   FLATPAK_PULL_FLAGS_ALLOW_DOWNGRADE = 1 << 2,
+  FLATPAK_PULL_FLAGS_NO_STATIC_DELTAS = 1 << 3,
 } FlatpakPullFlags;
 
 typedef enum {
@@ -106,10 +135,24 @@ typedef enum {
   FLATPAK_DIR_STORAGE_TYPE_HARD_DISK,
   FLATPAK_DIR_STORAGE_TYPE_SDCARD,
   FLATPAK_DIR_STORAGE_TYPE_MMC,
+  FLATPAK_DIR_STORAGE_TYPE_NETWORK,
 } FlatpakDirStorageType;
 
 GQuark       flatpak_dir_error_quark (void);
 
+#ifndef FLATPAK_ENABLE_P2P
+/* Rather than putting #ifdefs around all the function arguments for result sets,
+ * define away OstreeRepoFinderResult if we’re compiling without P2P support.
+ * The surrounding code should always pass in NULL if P2P support is disabled. */
+typedef void OstreeRepoFinderResult;
+typedef void** OstreeRepoFinderResultv;
+
+static inline void no_op (gpointer data) {}
+
+G_DEFINE_AUTOPTR_CLEANUP_FUNC (OstreeRepoFinderResult, no_op)
+G_DEFINE_AUTO_CLEANUP_FREE_FUNC (OstreeRepoFinderResultv, no_op, NULL)
+#endif  /* !FLATPAK_ENABLE_P2P */
+
 /**
  * FLATPAK_DEPLOY_DATA_GVARIANT_FORMAT:
  *
@@ -145,6 +188,12 @@ guint64             flatpak_deploy_data_get_installed_size (GVariant *deploy_dat
 const char *        flatpak_deploy_data_get_alt_id (GVariant *deploy_data);
 
 GFile *        flatpak_deploy_get_dir (FlatpakDeploy *deploy);
+GVariant *     flatpak_load_deploy_data (GFile *deploy_dir,
+                                         GCancellable *cancellable,
+                                         GError      **error);
+GVariant *     flatpak_deploy_get_deploy_data (FlatpakDeploy *deploy,
+                                               GCancellable *cancellable,
+                                               GError      **error);
 GFile *        flatpak_deploy_get_files (FlatpakDeploy *deploy);
 FlatpakContext *flatpak_deploy_get_overrides (FlatpakDeploy *deploy);
 GKeyFile *     flatpak_deploy_get_metadata (FlatpakDeploy *deploy);
@@ -171,6 +220,9 @@ gint        flatpak_dir_get_priority (FlatpakDir *self);
 FlatpakDirStorageType flatpak_dir_get_storage_type (FlatpakDir *self);
 GFile *     flatpak_dir_get_deploy_dir (FlatpakDir *self,
                                         const char *ref);
+char *      flatpak_dir_get_deploy_subdir (FlatpakDir *self,
+                                           const char *checksum,
+                                           const char * const * subpaths);
 GFile *     flatpak_dir_get_unmaintained_extension_dir (FlatpakDir *self,
                                                         const char *name,
                                                         const char *arch,
@@ -224,6 +276,16 @@ char **     flatpak_dir_find_remote_refs (FlatpakDir   *self,
                                           FlatpakKinds  kinds,
                                           GCancellable *cancellable,
                                           GError      **error);
+char *      flatpak_dir_find_local_ref (FlatpakDir   *self,
+                                        const char   *remote,
+                                        const char   *name,
+                                        const char   *opt_branch,
+                                        const char   *opt_default_branch,
+                                        const char   *opt_arch,
+                                        FlatpakKinds  kinds,
+                                        FlatpakKinds *out_kind,
+                                        GCancellable *cancellable,
+                                        GError      **error);
 char *      flatpak_dir_find_installed_ref (FlatpakDir  *self,
                                             const char  *opt_name,
                                             const char  *opt_branch,
@@ -259,6 +321,16 @@ gboolean    flatpak_dir_recreate_repo (FlatpakDir   *self,
 gboolean    flatpak_dir_ensure_repo (FlatpakDir   *self,
                                      GCancellable *cancellable,
                                      GError      **error);
+gboolean    flatpak_dir_maybe_ensure_repo (FlatpakDir   *self,
+                                           GCancellable *cancellable,
+                                           GError      **error);
+char *      flatpak_dir_get_config (FlatpakDir *self,
+				    const char *key,
+				    GError    **error);
+gboolean    flatpak_dir_set_config (FlatpakDir *self,
+                                    const char *key,
+                                    const char *value,
+                                    GError    **error);
 gboolean    flatpak_dir_mark_changed (FlatpakDir *self,
                                       GError    **error);
 gboolean    flatpak_dir_remove_appstream (FlatpakDir   *self,
@@ -271,6 +343,16 @@ gboolean    flatpak_dir_deploy_appstream (FlatpakDir          *self,
                                           gboolean            *out_changed,
                                           GCancellable        *cancellable,
                                           GError             **error);
+gboolean    flatpak_dir_find_latest_rev (FlatpakDir               *self,
+                                         const char               *remote,
+                                         const char               *ref,
+                                         char                    **out_rev,
+                                         OstreeRepoFinderResult ***out_results,
+                                         GCancellable             *cancellable,
+                                         GError                  **error);
+gboolean    flatpak_dir_check_for_appstream_update (FlatpakDir          *self,
+                                                    const char          *remote,
+                                                    const char          *arch);
 gboolean    flatpak_dir_update_appstream (FlatpakDir          *self,
                                           const char          *remote,
                                           const char          *arch,
@@ -282,6 +364,7 @@ gboolean    flatpak_dir_pull (FlatpakDir          *self,
                               const char          *repository,
                               const char          *ref,
                               const char          *opt_rev,
+                              const OstreeRepoFinderResult * const *results,
                               const char         **subpaths,
                               OstreeRepo          *repo,
                               FlatpakPullFlags     flatpak_flags,
@@ -342,6 +425,11 @@ gboolean    flatpak_dir_lock (FlatpakDir   *self,
                               GLnxLockFile *lockfile,
                               GCancellable *cancellable,
                               GError      **error);
+gboolean    flatpak_dir_repo_lock (FlatpakDir   *self,
+                                   GLnxLockFile *lockfile,
+                                   gboolean      exclusive,
+                                   GCancellable *cancellable,
+                                   GError      **error);
 gboolean    flatpak_dir_deploy (FlatpakDir          *self,
                                 const char          *origin,
                                 const char          *ref,
@@ -360,11 +448,14 @@ gboolean   flatpak_dir_deploy_install (FlatpakDir   *self,
                                        const char   *ref,
                                        const char   *origin,
                                        const char  **subpaths,
+                                       gboolean      reinstall,
                                        GCancellable *cancellable,
                                        GError      **error);
 gboolean   flatpak_dir_install (FlatpakDir          *self,
                                 gboolean             no_pull,
                                 gboolean             no_deploy,
+                                gboolean             no_static_deltas,
+                                gboolean             reinstall,
                                 const char          *ref,
                                 const char          *remote_name,
                                 const char         **subpaths,
@@ -385,25 +476,28 @@ gboolean flatpak_dir_install_bundle (FlatpakDir          *self,
                                      char               **out_ref,
                                      GCancellable        *cancellable,
                                      GError             **error);
+char * flatpak_dir_check_for_update (FlatpakDir          *self,
+                                     const char          *ref,
+                                     const char          *remote_name,
+                                     const char          *checksum_or_latest,
+                                     const char         **opt_subpaths,
+                                     gboolean             no_pull,
+                                     OstreeRepoFinderResult ***out_results,
+                                     GCancellable        *cancellable,
+                                     GError             **error);
 gboolean   flatpak_dir_update (FlatpakDir          *self,
                                gboolean             no_pull,
                                gboolean             no_deploy,
+                               gboolean             no_static_deltas,
+                               gboolean             allow_downgrade,
                                const char          *ref,
                                const char          *remote_name,
                                const char          *checksum_or_latest,
+                               const OstreeRepoFinderResult * const *results,
                                const char         **opt_subpaths,
                                OstreeAsyncProgress *progress,
                                GCancellable        *cancellable,
                                GError             **error);
-gboolean flatpak_dir_install_or_update (FlatpakDir          *self,
-                                        gboolean             no_pull,
-                                        gboolean             no_deploy,
-                                        const char          *ref,
-                                        const char          *remote_name,
-                                        const char         **opt_subpaths,
-                                        OstreeAsyncProgress *progress,
-                                        GCancellable        *cancellable,
-                                        GError             **error);
 gboolean flatpak_dir_uninstall (FlatpakDir          *self,
                                 const char          *ref,
                                 FlatpakHelperUninstallFlags flags,
@@ -437,6 +531,9 @@ gboolean    flatpak_dir_prune (FlatpakDir   *self,
 gboolean    flatpak_dir_cleanup_removed (FlatpakDir   *self,
                                          GCancellable *cancellable,
                                          GError      **error);
+gboolean    flatpak_dir_cleanup_undeployed_refs (FlatpakDir   *self,
+                                                 GCancellable *cancellable,
+                                                 GError      **error);
 gboolean    flatpak_dir_collect_deployed_refs (FlatpakDir   *self,
                                                const char   *type,
                                                const char   *name_prefix,
@@ -459,26 +556,34 @@ char      *flatpak_dir_create_origin_remote (FlatpakDir   *self,
                                              const char   *id,
                                              const char   *title,
                                              const char   *main_ref,
-                                             const char   *oci_uri,
-                                             const char   *oci_tag,
                                              GBytes       *gpg_data,
+                                             const char   *collection_id,
                                              GCancellable *cancellable,
                                              GError      **error);
+void       flatpak_dir_prune_origin_remote (FlatpakDir *self,
+                                            const char *remote);
 gboolean   flatpak_dir_create_remote_for_ref_file (FlatpakDir   *self,
                                                    GBytes  *data,
                                                    const char *default_arch,
                                                    char   **remote_name_out,
                                                    char   **ref_out,
                                                    GError **error);
+gboolean   flatpak_dir_create_suggested_remote_for_ref_file (FlatpakDir   *self,
+                                                             GBytes  *data,
+                                                             GError **error);
 GKeyFile * flatpak_dir_parse_repofile (FlatpakDir   *self,
                                        const char   *remote_name,
+                                       gboolean      from_ref,
                                        GBytes       *data,
                                        GBytes      **gpg_data_out,
                                        GCancellable *cancellable,
                                        GError      **error);
 
 char      *flatpak_dir_find_remote_by_uri (FlatpakDir   *self,
-                                           const char   *uri);
+                                           const char   *uri,
+                                           const char   *collection_id);
+gboolean   flatpak_dir_has_remote (FlatpakDir   *self,
+                                   const char   *remote_name);
 char     **flatpak_dir_list_remotes (FlatpakDir   *self,
                                      GCancellable *cancellable,
                                      GError      **error);
@@ -498,12 +603,12 @@ gboolean   flatpak_dir_remove_remote (FlatpakDir   *self,
                                       GError      **error);
 char      *flatpak_dir_get_remote_title (FlatpakDir *self,
                                          const char *remote_name);
+char      *flatpak_dir_get_remote_collection_id (FlatpakDir *self,
+                                                 const char *remote_name);
 char      *flatpak_dir_get_remote_main_ref (FlatpakDir *self,
                                             const char *remote_name);
-char      *flatpak_dir_get_remote_oci_uri (FlatpakDir *self,
-                                           const char *remote_name);
-char      *flatpak_dir_get_remote_oci_tag (FlatpakDir *self,
-                                           const char *remote_name);
+gboolean   flatpak_dir_get_remote_oci (FlatpakDir *self,
+                                       const char *remote_name);
 char      *flatpak_dir_get_remote_default_branch (FlatpakDir *self,
                                                   const char *remote_name);
 int        flatpak_dir_get_remote_prio (FlatpakDir *self,
@@ -519,6 +624,10 @@ gboolean   flatpak_dir_list_remote_refs (FlatpakDir   *self,
                                          GHashTable  **refs,
                                          GCancellable *cancellable,
                                          GError      **error);
+gboolean flatpak_dir_fetch_remote_repo_metadata (FlatpakDir    *self,
+                                                 const char    *remote_name,
+                                                 GCancellable  *cancellable,
+                                                 GError       **error);
 char *   flatpak_dir_fetch_remote_title (FlatpakDir   *self,
                                          const char   *remote,
                                          GCancellable *cancellable,
@@ -527,10 +636,31 @@ char *   flatpak_dir_fetch_remote_default_branch (FlatpakDir   *self,
                                                   const char   *remote,
                                                   GCancellable *cancellable,
                                                   GError      **error);
+GVariant * flatpak_dir_fetch_remote_commit (FlatpakDir   *self,
+                                            const char   *remote_name,
+                                            const char   *ref,
+                                            const char   *opt_commit,
+                                            char        **out_commit,
+                                            GCancellable *cancellable,
+                                            GError      **error);
 gboolean flatpak_dir_update_remote_configuration (FlatpakDir   *self,
                                                   const char   *remote,
                                                   GCancellable *cancellable,
                                                   GError      **error);
+gboolean flatpak_dir_update_remote_configuration_for_summary (FlatpakDir   *self,
+                                                              const char   *remote,
+                                                              GVariant     *summary,
+                                                              gboolean      dry_run,
+                                                              gboolean     *has_changed_out,
+                                                              GCancellable *cancellable,
+                                                              GError      **error);
+gboolean flatpak_dir_update_remote_configuration_for_repo_metadata (FlatpakDir    *self,
+                                                                    const char    *remote,
+                                                                    GVariant      *summary,
+                                                                    gboolean       dry_run,
+                                                                    gboolean      *has_changed_out,
+                                                                    GCancellable  *cancellable,
+                                                                    GError       **error);
 gboolean flatpak_dir_fetch_ref_cache (FlatpakDir   *self,
                                       const char   *remote_name,
                                       const char   *ref,
@@ -550,4 +680,16 @@ GPtrArray * flatpak_dir_find_local_related (FlatpakDir *self,
                                             GCancellable *cancellable,
                                             GError **error);
 
+gboolean flatpak_dir_lookup_repo_metadata (FlatpakDir    *self,
+                                           const char    *remote_name,
+                                           GCancellable  *cancellable,
+                                           GError       **error,
+                                           const char    *key,
+                                           const char    *format_string,
+                                           ...);
+
+char ** flatpak_dir_get_default_locale_languages (FlatpakDir *self);
+char ** flatpak_dir_get_locale_languages (FlatpakDir *self);
+char ** flatpak_dir_get_locale_subpaths (FlatpakDir *self);
+
 #endif /* __FLATPAK_DIR_H__ */
