@@ -73,8 +73,11 @@
  *    You can call the GetXXX methods on the name/id to get e.g. the peer pid
  *    You get AccessDenied rather than NameHasNoOwner when sending messages to the name/id
  *
+ * FILTERED:
+ *    You can send *some* method calls to the name (not id)
+ *
  * TALK:
- *    You can send method calls and signals to the name/id
+ *    You can send any method calls and signals to the name/id
  *    You will receive broadcast signals from the name/id (if you have a match rule for them)
  *    You can call StartServiceByName on the name
  *
@@ -170,9 +173,6 @@
 
 typedef struct FlatpakProxyClient FlatpakProxyClient;
 
-FlatpakPolicy flatpak_proxy_get_policy (FlatpakProxy *proxy,
-                                        const char   *name);
-
 #define FIND_AUTH_END_CONTINUE -1
 #define FIND_AUTH_END_ABORT -2
 
@@ -222,6 +222,13 @@ typedef struct
   guint32     unix_fds;
 } Header;
 
+typedef struct
+{
+  char *path;
+  char *interface;
+  char *member;
+} Filter;
+
 static void header_free (Header *header);
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (Header, header_free)
 
@@ -283,9 +290,11 @@ struct FlatpakProxy
   char          *dbus_address;
 
   gboolean       filter;
+  gboolean       sloppy_names;
 
   GHashTable    *wildcard_policy;
   GHashTable    *policy;
+  GHashTable    *filters;
 };
 
 typedef struct
@@ -406,7 +415,7 @@ flatpak_proxy_client_init (FlatpakProxyClient *client)
   client->unique_id_policy = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
 }
 
-FlatpakProxyClient *
+static FlatpakProxyClient *
 flatpak_proxy_client_new (FlatpakProxy *proxy, GSocketConnection *connection)
 {
   FlatpakProxyClient *client;
@@ -442,7 +451,7 @@ flatpak_proxy_get_wildcard_policy (FlatpakProxy *proxy,
   return wildcard_policy;
 }
 
-FlatpakPolicy
+static FlatpakPolicy
 flatpak_proxy_get_policy (FlatpakProxy *proxy,
                           const char   *name)
 {
@@ -462,6 +471,13 @@ flatpak_proxy_set_filter (FlatpakProxy *proxy,
   proxy->filter = filter;
 }
 
+void
+flatpak_proxy_set_sloppy_names (FlatpakProxy *proxy,
+                                gboolean      sloppy_names)
+{
+  proxy->sloppy_names = sloppy_names;
+}
+
 void
 flatpak_proxy_set_log_messages (FlatpakProxy *proxy,
                                 gboolean      log)
@@ -474,7 +490,11 @@ flatpak_proxy_add_policy (FlatpakProxy *proxy,
                           const char   *name,
                           FlatpakPolicy policy)
 {
-  g_hash_table_replace (proxy->policy, g_strdup (name), GINT_TO_POINTER (policy));
+  guint current_policy = GPOINTER_TO_INT (g_hash_table_lookup (proxy->policy, name));
+
+  current_policy = MAX (policy, current_policy);
+
+  g_hash_table_replace (proxy->policy, g_strdup (name), GINT_TO_POINTER (current_policy));
 }
 
 void
@@ -485,6 +505,77 @@ flatpak_proxy_add_wildcarded_policy (FlatpakProxy *proxy,
   g_hash_table_replace (proxy->wildcard_policy, g_strdup (name), GINT_TO_POINTER (policy));
 }
 
+static void
+filter_free (Filter *filter)
+{
+  g_free (filter->path);
+  g_free (filter->interface);
+  g_free (filter->member);
+  g_free (filter);
+}
+
+static void
+filter_list_free (GList *filters)
+{
+  g_list_free_full (filters, (GDestroyNotify)filter_free);
+}
+
+/* rules are of the form [org.the.interface.[method|*]][@/obj/path] */
+static Filter *
+filter_new (const char *rule)
+{
+  Filter *filter = g_new0 (Filter, 1);
+  const char *obj_path_start = NULL;
+  const char *method_end = NULL;
+
+  obj_path_start = strchr (rule, '@');
+  if (obj_path_start && obj_path_start[1] != 0)
+    filter->path = g_strdup (obj_path_start + 1);
+
+  if (obj_path_start != NULL)
+    method_end = obj_path_start;
+  else
+    method_end = rule + strlen(rule);
+
+  if (rule[0] != '@')
+    {
+      filter->interface = g_strndup (rule, method_end - rule);
+      char *dot = strrchr (filter->interface, '.');
+      if (dot != NULL)
+        {
+          *dot = 0;
+          if (strcmp (dot+1, "*") != 0)
+            filter->member = g_strdup (dot + 1);
+         }
+    }
+
+  return filter;
+}
+
+
+void
+flatpak_proxy_add_filter (FlatpakProxy *proxy,
+                          const char   *name,
+                          const char   *rule)
+{
+  Filter *filter;
+  GList *filters, *new_filters;
+
+  filter = filter_new (rule);
+  if (g_hash_table_lookup_extended (proxy->filters,
+                                    name,
+                                    NULL, (void **)&filters))
+    {
+      new_filters = g_list_append (filters, filter);
+      g_assert (new_filters == filters);
+    }
+  else
+    {
+      filters = g_list_append (NULL, filter);
+      g_hash_table_insert (proxy->filters, g_strdup (name), filters);
+    }
+}
+
 static void
 flatpak_proxy_finalize (GObject *object)
 {
@@ -493,11 +584,11 @@ flatpak_proxy_finalize (GObject *object)
   if (g_socket_service_is_active (G_SOCKET_SERVICE (proxy)))
     unlink (proxy->socket_path);
 
-  g_clear_pointer (&proxy->dbus_address, g_free);
   g_assert (proxy->clients == NULL);
 
   g_hash_table_destroy (proxy->policy);
   g_hash_table_destroy (proxy->wildcard_policy);
+  g_hash_table_destroy (proxy->filters);
 
   g_free (proxy->socket_path);
   g_free (proxy->dbus_address);
@@ -1406,9 +1497,29 @@ get_dbus_method_handler (FlatpakProxyClient *client, Header *header)
   policy = flatpak_proxy_client_get_policy (client, header->destination);
   if (policy < FLATPAK_POLICY_SEE)
     return HANDLE_HIDE;
-  if (policy < FLATPAK_POLICY_TALK)
+  if (policy < FLATPAK_POLICY_FILTERED)
     return HANDLE_DENY;
 
+  if (policy == FLATPAK_POLICY_FILTERED)
+    {
+      GList *filters = NULL, *l;
+
+      if (header->destination)
+        filters = g_hash_table_lookup (client->proxy->filters, header->destination);
+      for (l = filters; l != NULL; l = l->next)
+        {
+          Filter *filter = l->data;
+
+          if (header->type == G_DBUS_MESSAGE_TYPE_METHOD_CALL &&
+              (filter->path == NULL || g_strcmp0 (filter->path, header->path) == 0) &&
+              (filter->interface == NULL || g_strcmp0 (filter->interface, header->interface) == 0) &&
+              (filter->member == NULL || g_strcmp0 (filter->member, header->member) == 0))
+            return HANDLE_PASS;
+        }
+
+      return HANDLE_DENY;
+    }
+
   if (!is_for_bus (header))
     return HANDLE_PASS;
 
@@ -1459,7 +1570,7 @@ get_dbus_method_handler (FlatpakProxyClient *client, Header *header)
           strcmp (method, "ListActivatableNames") == 0)
         return HANDLE_FILTER_NAME_LIST_REPLY;
 
-      g_warning ("Unknown bus method %s\n", method);
+      g_warning ("Unknown bus method %s", method);
       return HANDLE_DENY;
     }
   else
@@ -1629,7 +1740,8 @@ should_filter_name_owner_changed (FlatpakProxyClient *client, Buffer *buffer)
   old = g_variant_get_string (arg1, NULL);
   new = g_variant_get_string (arg2, NULL);
 
-  if (flatpak_proxy_client_get_policy (client, name) >= FLATPAK_POLICY_SEE)
+  if (flatpak_proxy_client_get_policy (client, name) >= FLATPAK_POLICY_SEE ||
+      (client->proxy->sloppy_names && name[0] == ':'))
     {
       if (name[0] != ':')
         {
@@ -1948,10 +2060,10 @@ got_buffer_from_client (FlatpakProxyClient *client, ProxySide *side, Buffer *buf
         case HANDLE_PASS:
 handle_pass:
           if (client_message_generates_reply (header))
-	    {
-	      if (expecting_reply == EXPECTED_REPLY_NONE)
-		expecting_reply = EXPECTED_REPLY_NORMAL;
-	    }
+            {
+              if (expecting_reply == EXPECTED_REPLY_NONE)
+                expecting_reply = EXPECTED_REPLY_NORMAL;
+            }
 
           break;
 
@@ -1973,7 +2085,7 @@ handle_hide:
                 error = "org.freedesktop.DBus.Error.ServiceUnknown";
 
               buffer = get_error_for_roundtrip (client, header, error);
-	      expecting_reply = EXPECTED_REPLY_REWRITE;
+              expecting_reply = EXPECTED_REPLY_REWRITE;
             }
           else
             {
@@ -1993,8 +2105,8 @@ handle_deny:
                 g_print ("*DENIED* (ping)\n");
 
               buffer = get_error_for_roundtrip (client, header,
-						"org.freedesktop.DBus.Error.AccessDenied");
-	      expecting_reply = EXPECTED_REPLY_REWRITE;
+                                                "org.freedesktop.DBus.Error.AccessDenied");
+              expecting_reply = EXPECTED_REPLY_REWRITE;
             }
           else
             {
@@ -2146,7 +2258,7 @@ got_buffer_from_bus (FlatpakProxyClient *client, ProxySide *side, Buffer *buffer
               break;
 
             default:
-              g_warning ("Unexpected expected reply type %d\n", expected_reply);
+              g_warning ("Unexpected expected reply type %d", expected_reply);
             }
         }
       else /* Not reply */
@@ -2162,12 +2274,12 @@ got_buffer_from_bus (FlatpakProxyClient *client, ProxySide *side, Buffer *buffer
             }
 
           /* We filter all NameOwnerChanged signal according to the policy */
-	  if (message_is_name_owner_changed (client, header))
-	    {
-	      if (should_filter_name_owner_changed (client, buffer))
-		g_clear_pointer (&buffer, buffer_unref);
-	    }
-	}
+          if (message_is_name_owner_changed (client, header))
+            {
+              if (should_filter_name_owner_changed (client, buffer))
+                g_clear_pointer (&buffer, buffer_unref);
+            }
+        }
 
       /* All incoming broadcast signals are filtered according to policy */
       if (header->type == G_DBUS_MESSAGE_TYPE_SIGNAL && header->destination == NULL)
@@ -2441,7 +2553,7 @@ client_connected_to_dbus (GObject      *source_object,
   stream = g_dbus_address_get_stream_finish (res, NULL, &error);
   if (stream == NULL)
     {
-      g_warning ("Failed to connect to bus: %s\n", error->message);
+      g_warning ("Failed to connect to bus: %s", error->message);
       g_object_unref (client);
       return;
     }
@@ -2475,6 +2587,7 @@ static void
 flatpak_proxy_init (FlatpakProxy *proxy)
 {
   proxy->policy = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+  proxy->filters = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, (GDestroyNotify)filter_list_free);
   proxy->wildcard_policy = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
   flatpak_proxy_add_policy (proxy, "org.freedesktop.DBus", FLATPAK_POLICY_TALK);
 }
