@@ -56,9 +56,10 @@ typedef struct
  * Flags to alter the behavior of flatpak_installation_update().
  */
 typedef enum {
-  FLATPAK_UPDATE_FLAGS_NONE      = 0,
-  FLATPAK_UPDATE_FLAGS_NO_DEPLOY = (1 << 0),
-  FLATPAK_UPDATE_FLAGS_NO_PULL   = (1 << 1),
+  FLATPAK_UPDATE_FLAGS_NONE             = 0,
+  FLATPAK_UPDATE_FLAGS_NO_DEPLOY        = (1 << 0),
+  FLATPAK_UPDATE_FLAGS_NO_PULL          = (1 << 1),
+  FLATPAK_UPDATE_FLAGS_NO_STATIC_DELTAS = (1 << 2),
 } FlatpakUpdateFlags;
 
 /**
@@ -68,7 +69,10 @@ typedef enum {
  * Flags to alter the behavior of flatpak_installation_install_full().
  */
 typedef enum {
-  FLATPAK_INSTALL_FLAGS_NONE      = 0,
+  FLATPAK_INSTALL_FLAGS_NONE             = 0,
+  FLATPAK_INSTALL_FLAGS_NO_STATIC_DELTAS = (1 << 0),
+  FLATPAK_INSTALL_FLAGS_NO_DEPLOY        = (1 << 2),
+  FLATPAK_INSTALL_FLAGS_NO_PULL          = (1 << 3),
 } FlatpakInstallFlags;
 
 /**
@@ -77,6 +81,7 @@ typedef enum {
  * @FLATPAK_STORAGE_TYPE_HARD_DISK: installation is on a hard disk
  * @FLATPAK_STORAGE_TYPE_SDCARD: installation is on a SD card
  * @FLATPAK_STORAGE_TYPE_MMC: installation is on an MMC
+ * @FLATPAK_STORAGE_TYPE_NETWORK: installation is on the network
  *
  * Flags to alter the behavior of flatpak_installation_install_full().
  *
@@ -87,6 +92,7 @@ typedef enum {
   FLATPAK_STORAGE_TYPE_HARD_DISK,
   FLATPAK_STORAGE_TYPE_SDCARD,
   FLATPAK_STORAGE_TYPE_MMC,
+  FLATPAK_STORAGE_TYPE_NETWORK,
 } FlatpakStorageType;
 
 
@@ -189,6 +195,15 @@ FLATPAK_EXTERN gboolean              flatpak_installation_update_remote_sync (Fl
                                                                               const char          *name,
                                                                               GCancellable        *cancellable,
                                                                               GError             **error);
+FLATPAK_EXTERN gboolean              flatpak_installation_set_config_sync (FlatpakInstallation *self,
+									   const char          *key,
+									   const char          *value,
+									   GCancellable        *cancellable,
+									   GError             **error);
+FLATPAK_EXTERN char *                flatpak_installation_get_config      (FlatpakInstallation *self,
+									   const char          *key,
+									   GCancellable        *cancellable,
+									   GError             **error);
 FLATPAK_EXTERN char *              flatpak_installation_load_app_overrides (FlatpakInstallation *self,
                                                                             const char          *app_id,
                                                                             GCancellable        *cancellable,
@@ -286,6 +301,14 @@ FLATPAK_EXTERN gboolean          flatpak_installation_update_appstream_sync (Fla
                                                                              gboolean            *out_changed,
                                                                              GCancellable        *cancellable,
                                                                              GError             **error);
+FLATPAK_EXTERN gboolean          flatpak_installation_update_appstream_full_sync (FlatpakInstallation *self,
+                                                                                  const char          *remote_name,
+                                                                                  const char          *arch,
+                                                                                  FlatpakProgressCallback progress,
+                                                                                  gpointer                progress_data,
+                                                                                  gboolean            *out_changed,
+                                                                                  GCancellable        *cancellable,
+                                                                                  GError             **error);
 FLATPAK_EXTERN GPtrArray    *    flatpak_installation_list_remote_related_refs_sync (FlatpakInstallation *self,
                                                                                      const char          *remote_name,
                                                                                      const char          *ref,
@@ -297,4 +320,17 @@ FLATPAK_EXTERN GPtrArray    *    flatpak_installation_list_installed_related_ref
                                                                                         GCancellable        *cancellable,
                                                                                         GError             **error);
 
+FLATPAK_EXTERN gboolean          flatpak_installation_remove_local_ref_sync (FlatpakInstallation *self,
+                                                                             const char          *remote_name,
+                                                                             const char          *ref,
+                                                                             GCancellable        *cancellable,
+                                                                             GError              **error);
+FLATPAK_EXTERN gboolean          flatpak_installation_cleanup_local_refs_sync (FlatpakInstallation *self,
+                                                                               GCancellable        *cancellable,
+                                                                               GError              **error);
+FLATPAK_EXTERN gboolean          flatpak_installation_prune_local_repo (FlatpakInstallation *self,
+                                                                        GCancellable        *cancellable,
+                                                                        GError              **error);
+
+
 #endif /* __FLATPAK_INSTALLATION_H__ */
