@@ -28,6 +28,7 @@
 
 #include "flatpak-dbus.h"
 #include "flatpak-dir.h"
+#include "flatpak-oci-registry.h"
 #include "lib/flatpak-error.h"
 
 static PolkitAuthority *authority = NULL;
@@ -134,6 +135,11 @@ dir_get_system (const char *installation, GError **error)
   return system;
 }
 
+static void
+no_progress_cb (OstreeAsyncProgress *progress, gpointer user_data)
+{
+}
+
 static gboolean
 handle_deploy (FlatpakSystemHelper   *object,
                GDBusMethodInvocation *invocation,
@@ -148,9 +154,12 @@ handle_deploy (FlatpakSystemHelper   *object,
   g_autoptr(GFile) path = g_file_new_for_path (arg_repo_path);
   g_autoptr(GError) error = NULL;
   g_autoptr(GFile) deploy_dir = NULL;
+  g_autoptr(OstreeAsyncProgress) ostree_progress = NULL;
   gboolean is_update;
+  gboolean is_oci;
   gboolean no_deploy;
   gboolean local_pull;
+  gboolean reinstall;
   g_autoptr(GMainContext) main_context = NULL;
   g_autofree char *url = NULL;
 
@@ -179,27 +188,32 @@ handle_deploy (FlatpakSystemHelper   *object,
   is_update = (arg_flags & FLATPAK_HELPER_DEPLOY_FLAGS_UPDATE) != 0;
   no_deploy = (arg_flags & FLATPAK_HELPER_DEPLOY_FLAGS_NO_DEPLOY) != 0;
   local_pull = (arg_flags & FLATPAK_HELPER_DEPLOY_FLAGS_LOCAL_PULL) != 0;
+  reinstall = (arg_flags & FLATPAK_HELPER_DEPLOY_FLAGS_REINSTALL) != 0;
 
-  deploy_dir = flatpak_dir_get_if_deployed (system, arg_ref,
-                                            NULL, NULL);
+  deploy_dir = flatpak_dir_get_if_deployed (system, arg_ref, NULL, NULL);
 
   if (deploy_dir)
     {
       g_autofree char *real_origin = NULL;
       if (!is_update)
         {
-          /* Can't install already installed app */
-          g_dbus_method_invocation_return_error (invocation, FLATPAK_ERROR, FLATPAK_ERROR_ALREADY_INSTALLED,
-                                                 "%s is already installed", arg_ref);
-          return TRUE;
+          if (!reinstall)
+            {
+              /* Can't install already installed app */
+              g_dbus_method_invocation_return_error (invocation, FLATPAK_ERROR, FLATPAK_ERROR_ALREADY_INSTALLED,
+                                                     "%s is already installed", arg_ref);
+              return TRUE;
+            }
         }
-
-      real_origin = flatpak_dir_get_origin (system, arg_ref, NULL, NULL);
-      if (g_strcmp0 (real_origin, arg_origin) != 0)
+      else
         {
-          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
-                                                 "Wrong origin %s for update", arg_origin);
-          return TRUE;
+          real_origin = flatpak_dir_get_origin (system, arg_ref, NULL, NULL);
+          if (g_strcmp0 (real_origin, arg_origin) != 0)
+            {
+              g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
+                                                     "Wrong origin %s for update", arg_origin);
+              return TRUE;
+            }
         }
     }
   else if (!deploy_dir && is_update)
@@ -217,17 +231,98 @@ handle_deploy (FlatpakSystemHelper   *object,
       return TRUE;
     }
 
-  if (strlen (arg_repo_path) > 0)
+  is_oci = flatpak_dir_get_remote_oci (system, arg_origin);
+
+  if (strlen (arg_repo_path) > 0 && is_oci)
+    {
+      g_autoptr(GFile) registry_file = g_file_new_for_path (arg_repo_path);
+      g_autofree char *registry_uri = g_file_get_uri (registry_file);
+      g_autoptr(FlatpakOciRegistry) registry = NULL;
+      g_autoptr(FlatpakOciIndex) index = NULL;
+      const FlatpakOciManifestDescriptor *desc;
+      g_autoptr(FlatpakOciVersioned) versioned = NULL;
+      g_autofree char *checksum = NULL;
+      g_autofree char *upstream_url = NULL;
+      g_autoptr(SoupSession) soup_session = NULL;
+
+      ostree_repo_remote_get_url (flatpak_dir_get_repo (system),
+                                  arg_origin,
+                                  &upstream_url,
+                                  NULL);
+
+      if (upstream_url == NULL)
+	{
+	  g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+						 "Remote %s is disabled", arg_origin);
+	  return TRUE;
+	}
+
+      registry = flatpak_oci_registry_new (registry_uri, FALSE, -1, NULL, &error);
+      if (registry == NULL)
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Can't open child OCI registry: %s", error->message);
+          return TRUE;
+        }
+
+      index = flatpak_oci_registry_load_index (registry, NULL, NULL, NULL, &error);
+      if (index == NULL)
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Can't open child OCI registry index: %s", error->message);
+          return TRUE;
+        }
+
+      desc = flatpak_oci_index_get_manifest (index, arg_ref);
+      if (desc == NULL)
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Can't find ref %s in child OCI registry index", arg_ref);
+          return TRUE;
+        }
+
+      versioned = flatpak_oci_registry_load_versioned (registry, NULL, desc->parent.digest, NULL,
+                                                       NULL, &error);
+      if (versioned == NULL || !FLATPAK_IS_OCI_MANIFEST (versioned))
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Can't open child manifest");
+          return TRUE;
+        }
+
+      soup_session = flatpak_create_soup_session (PACKAGE_STRING);
+      if (!flatpak_oci_index_verify_ref (soup_session,
+					 upstream_url,
+					 arg_ref,
+					 desc->parent.digest,
+					 NULL, &error))
+	{
+	  g_dbus_method_invocation_return_gerror (invocation, error);
+	  return TRUE;
+	}
+
+      checksum = flatpak_pull_from_oci (flatpak_dir_get_repo (system), registry, NULL, desc->parent.digest, FLATPAK_OCI_MANIFEST (versioned),
+                                        arg_origin, arg_ref, NULL, NULL, NULL, &error);
+      if (checksum == NULL)
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Can't pull ref %s from child OCI registry index: %s", arg_ref, error->message);
+          return TRUE;
+        }
+    }
+  else if (strlen (arg_repo_path) > 0)
     {
       /* Work around ostree-pull spinning the default main context for the sync calls */
       main_context = g_main_context_new ();
       g_main_context_push_thread_default (main_context);
 
+      ostree_progress = ostree_async_progress_new_and_connect (no_progress_cb, NULL);
+
       if (!flatpak_dir_pull_untrusted_local (system, arg_repo_path,
                                              arg_origin,
                                              arg_ref,
                                              (const char **) arg_subpaths,
-                                             NULL,
+                                             ostree_progress,
                                              NULL, &error))
         {
           g_main_context_pop_thread_default (main_context);
@@ -236,6 +331,9 @@ handle_deploy (FlatpakSystemHelper   *object,
           return TRUE;
         }
       g_main_context_pop_thread_default (main_context);
+
+      if (ostree_progress)
+        ostree_async_progress_finish (ostree_progress);
     }
   else if (local_pull)
     {
@@ -260,8 +358,10 @@ handle_deploy (FlatpakSystemHelper   *object,
       main_context = g_main_context_new ();
       g_main_context_push_thread_default (main_context);
 
-      if (!flatpak_dir_pull (system, arg_origin, arg_ref, NULL, (const char **)arg_subpaths, NULL,
-                             FLATPAK_PULL_FLAGS_NONE, OSTREE_REPO_PULL_FLAGS_UNTRUSTED, NULL,
+      ostree_progress = ostree_async_progress_new_and_connect (no_progress_cb, NULL);
+
+      if (!flatpak_dir_pull (system, arg_origin, arg_ref, NULL, NULL, (const char **)arg_subpaths, NULL,
+                             FLATPAK_PULL_FLAGS_NONE, OSTREE_REPO_PULL_FLAGS_UNTRUSTED, ostree_progress,
                              NULL, &error))
         {
           g_main_context_pop_thread_default (main_context);
@@ -269,7 +369,11 @@ handle_deploy (FlatpakSystemHelper   *object,
                                                  "Error pulling from repo: %s", error->message);
           return TRUE;
         }
+
       g_main_context_pop_thread_default (main_context);
+
+      if (ostree_progress)
+        ostree_async_progress_finish (ostree_progress);
     }
 
   if (!no_deploy)
@@ -292,6 +396,7 @@ handle_deploy (FlatpakSystemHelper   *object,
         {
           if (!flatpak_dir_deploy_install (system, arg_ref, arg_origin,
                                            (const char **) arg_subpaths,
+                                           reinstall,
                                            NULL, &error))
             {
               g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
@@ -319,6 +424,7 @@ handle_deploy_appstream (FlatpakSystemHelper   *object,
   g_autoptr(GError) error = NULL;
   g_autoptr(GMainContext) main_context = NULL;
   g_autofree char *branch = NULL;
+  gboolean is_oci;
 
   g_debug ("DeployAppstream %s %s %s %s", arg_repo_path, arg_origin, arg_arch, arg_installation);
 
@@ -342,26 +448,83 @@ handle_deploy_appstream (FlatpakSystemHelper   *object,
       return TRUE;
     }
 
-  /* Work around ostree-pull spinning the default main context for the sync calls */
-  main_context = g_main_context_new ();
-  g_main_context_push_thread_default (main_context);
+  is_oci = flatpak_dir_get_remote_oci (system, arg_origin);
 
   branch = g_strdup_printf ("appstream/%s", arg_arch);
 
-  if (!flatpak_dir_pull_untrusted_local (system, arg_repo_path,
-                                         arg_origin,
-                                         branch,
-                                         NULL,
-                                         NULL,
-                                         NULL, &error))
+  if (is_oci)
     {
-      g_main_context_pop_thread_default (main_context);
-      g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
-                                             "Error pulling from repo: %s", error->message);
-      return TRUE;
-    }
+      g_autoptr(GFile) registry_file = g_file_new_for_path (arg_repo_path);
+      g_autofree char *registry_uri = g_file_get_uri (registry_file);
+      g_autoptr(FlatpakOciRegistry) registry = NULL;
+      g_autoptr(FlatpakOciIndex) index = NULL;
+      const FlatpakOciManifestDescriptor *desc;
+      g_autoptr(FlatpakOciVersioned) versioned = NULL;
+      g_autofree char *checksum = NULL;
 
-  g_main_context_pop_thread_default (main_context);
+      registry = flatpak_oci_registry_new (registry_uri, FALSE, -1, NULL, &error);
+      if (registry == NULL)
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Can't open child OCI registry: %s", error->message);
+          return TRUE;
+        }
+
+      index = flatpak_oci_registry_load_index (registry, NULL, NULL, NULL, &error);
+      if (index == NULL)
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Can't open child OCI registry index: %s", error->message);
+          return TRUE;
+        }
+
+      desc = flatpak_oci_index_get_manifest (index, branch);
+      if (desc == NULL)
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Can't find ref %s in child OCI registry index", branch);
+          return TRUE;
+        }
+
+      versioned = flatpak_oci_registry_load_versioned (registry, NULL, desc->parent.digest, NULL,
+                                                       NULL, &error);
+      if (versioned == NULL || !FLATPAK_IS_OCI_MANIFEST (versioned))
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Can't open child manifest");
+          return TRUE;
+        }
+
+      checksum = flatpak_pull_from_oci (flatpak_dir_get_repo (system), registry, NULL, desc->parent.digest, FLATPAK_OCI_MANIFEST (versioned),
+                                        arg_origin, branch, NULL, NULL, NULL, &error);
+      if (checksum == NULL)
+        {
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Can't pull ref %s from child OCI registry index: %s", branch, error->message);
+          return TRUE;
+        }
+    }
+  else
+    {
+      /* Work around ostree-pull spinning the default main context for the sync calls */
+      main_context = g_main_context_new ();
+      g_main_context_push_thread_default (main_context);
+
+      if (!flatpak_dir_pull_untrusted_local (system, arg_repo_path,
+                                             arg_origin,
+                                             branch,
+                                             NULL,
+                                             NULL,
+                                             NULL, &error))
+        {
+          g_main_context_pop_thread_default (main_context);
+          g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
+                                                 "Error pulling from repo: %s", error->message);
+          return TRUE;
+        }
+
+      g_main_context_pop_thread_default (main_context);
+    }
 
   if (!flatpak_dir_deploy_appstream (system,
                                      arg_origin,
@@ -408,13 +571,13 @@ handle_uninstall (FlatpakSystemHelper *object,
 
   if (!flatpak_dir_ensure_repo (system, NULL, &error))
     {
-      g_dbus_method_invocation_return_gerror  (invocation, error);
+      g_dbus_method_invocation_return_gerror (invocation, error);
       return TRUE;
     }
 
   if (!flatpak_dir_uninstall (system, arg_ref, arg_flags, NULL, &error))
     {
-      g_dbus_method_invocation_return_gerror  (invocation, error);
+      g_dbus_method_invocation_return_gerror (invocation, error);
       return TRUE;
     }
 
@@ -461,7 +624,7 @@ handle_install_bundle (FlatpakSystemHelper   *object,
 
   if (!flatpak_dir_install_bundle (system, path, arg_remote, &ref, NULL, &error))
     {
-      g_dbus_method_invocation_return_gerror  (invocation, error);
+      g_dbus_method_invocation_return_gerror (invocation, error);
       return TRUE;
     }
 
@@ -520,7 +683,7 @@ handle_configure_remote (FlatpakSystemHelper *object,
 
   if (!flatpak_dir_ensure_repo (system, NULL, &error))
     {
-      g_dbus_method_invocation_return_gerror  (invocation, error);
+      g_dbus_method_invocation_return_gerror (invocation, error);
       return TRUE;
     }
 
@@ -535,7 +698,7 @@ handle_configure_remote (FlatpakSystemHelper *object,
       if (!flatpak_dir_modify_remote (system, arg_remote, config,
                                       gpg_data, NULL, &error))
         {
-          g_dbus_method_invocation_return_gerror  (invocation, error);
+          g_dbus_method_invocation_return_gerror (invocation, error);
           return TRUE;
         }
     }
@@ -547,7 +710,7 @@ handle_configure_remote (FlatpakSystemHelper *object,
                                       arg_remote,
                                       NULL, &error))
         {
-          g_dbus_method_invocation_return_gerror  (invocation, error);
+          g_dbus_method_invocation_return_gerror (invocation, error);
           return TRUE;
         }
     }
@@ -557,6 +720,251 @@ handle_configure_remote (FlatpakSystemHelper *object,
   return TRUE;
 }
 
+static gboolean
+handle_configure (FlatpakSystemHelper *object,
+		  GDBusMethodInvocation *invocation,
+		  guint arg_flags,
+		  const gchar *arg_key,
+		  const gchar *arg_value,
+		  const gchar *arg_installation)
+{
+  g_autoptr(FlatpakDir) system = NULL;
+  g_autoptr(GError) error = NULL;
+
+  g_debug ("Configure %u %s=%s %s", arg_flags, arg_key, arg_value, arg_installation);
+
+  system = dir_get_system (arg_installation, &error);
+  if (system == NULL)
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if ((arg_flags & ~FLATPAK_HELPER_CONFIGURE_FLAGS_ALL) != 0)
+    {
+      g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
+                                             "Unsupported flags enabled: 0x%x", (arg_flags & ~FLATPAK_HELPER_CONFIGURE_FLAGS_ALL));
+      return TRUE;
+    }
+
+  /* We only support this for now */
+  if (strcmp (arg_key, "languages") != 0)
+    {
+      g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
+                                             "Unsupported key: %s", arg_key);
+      return TRUE;
+    }
+
+  if ((arg_flags & FLATPAK_HELPER_CONFIGURE_FLAGS_UNSET) != 0)
+    arg_value = NULL;
+
+  if (!flatpak_dir_ensure_repo (system, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (!flatpak_dir_set_config (system, arg_key, arg_value, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  flatpak_system_helper_complete_configure (object, invocation);
+
+  return TRUE;
+}
+
+static gboolean
+handle_update_remote (FlatpakSystemHelper *object,
+                      GDBusMethodInvocation *invocation,
+                      guint arg_flags,
+                      const gchar *arg_remote,
+                      const gchar *arg_installation,
+                      const gchar *arg_summary_path,
+                      const gchar *arg_summary_sig_path)
+{
+  g_autoptr(FlatpakDir) system = NULL;
+  g_autoptr(GError) error = NULL;
+  char *summary_data = NULL;
+  gsize summary_size;
+  g_autoptr(GBytes) summary_bytes = NULL;
+  g_autoptr(GVariant) summary = NULL;
+  char *summary_sig_data = NULL;
+  gsize summary_sig_size;
+  g_autoptr(GBytes) summary_sig_bytes = NULL;
+  g_autoptr(OstreeGpgVerifyResult) gpg_result = NULL;
+  g_autofree char *collection_id = NULL;
+
+  g_debug ("UpdateRemote %u %s %s %s %s", arg_flags, arg_remote, arg_installation, arg_summary_path, arg_summary_sig_path);
+
+  system = dir_get_system (arg_installation, &error);
+  if (system == NULL)
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (*arg_remote == 0 || strchr (arg_remote, '/') != NULL)
+    {
+      g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
+                                             "Invalid remote name: %s", arg_remote);
+      return TRUE;
+    }
+
+  if ((arg_flags & ~FLATPAK_HELPER_UPDATE_REMOTE_FLAGS_ALL) != 0)
+    {
+      g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
+                                             "Unsupported flags enabled: 0x%x", (arg_flags & ~FLATPAK_HELPER_UPDATE_REMOTE_FLAGS_ALL));
+      return TRUE;
+    }
+
+  if (!flatpak_dir_ensure_repo (system, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+#ifdef FLATPAK_ENABLE_P2P
+  if (!ostree_repo_get_remote_option (flatpak_dir_get_repo (system), arg_remote, "collection-id",
+                                      NULL, &collection_id, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+  else if (collection_id != NULL && *collection_id == '\0')
+    g_clear_pointer (&collection_id, g_free);
+#endif  /* FLATPAK_ENABLE_P2P */
+
+  if (!g_file_get_contents (arg_summary_path, &summary_data, &summary_size, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+  summary_bytes = g_bytes_new_take (summary_data, summary_size);
+
+  if (collection_id == NULL)
+    {
+      if (!g_file_get_contents (arg_summary_sig_path, &summary_sig_data, &summary_sig_size, &error))
+        {
+          g_dbus_method_invocation_return_gerror (invocation, error);
+          return TRUE;
+        }
+      summary_sig_bytes = g_bytes_new_take (summary_sig_data, summary_sig_size);
+
+      gpg_result = ostree_repo_verify_summary (flatpak_dir_get_repo (system),
+                                               arg_remote,
+                                               summary_bytes,
+                                               summary_sig_bytes,
+                                               NULL, &error);
+      if (gpg_result == NULL ||
+          !ostree_gpg_verify_result_require_valid_signature (gpg_result, &error))
+        {
+          g_dbus_method_invocation_return_gerror (invocation, error);
+          return TRUE;
+        }
+
+      summary = g_variant_ref_sink (g_variant_new_from_bytes (OSTREE_SUMMARY_GVARIANT_FORMAT,
+                                                              summary_bytes, FALSE));
+      if (!flatpak_dir_update_remote_configuration_for_summary (system, arg_remote, summary,
+                                                                FALSE, NULL, NULL, &error))
+        {
+          g_dbus_method_invocation_return_gerror (invocation, error);
+          return TRUE;
+        }
+    }
+  else
+    {
+      if (!flatpak_dir_update_remote_configuration_for_repo_metadata (system, arg_remote, summary,
+                                                                      FALSE, NULL, NULL, &error))
+        {
+          g_dbus_method_invocation_return_gerror (invocation, error);
+          return TRUE;
+        }
+    }
+
+  flatpak_system_helper_complete_update_remote (object, invocation);
+
+  return TRUE;
+}
+
+static gboolean
+handle_remove_local_ref (FlatpakSystemHelper   *object,
+                         GDBusMethodInvocation *invocation,
+                         const gchar           *arg_remote,
+                         const gchar           *arg_ref,
+                         const gchar           *arg_installation)
+{
+  g_autoptr(FlatpakDir) system = NULL;
+  g_autoptr(GError) error = NULL;
+
+  g_debug ("RemoveLocalRef %s %s %s", arg_remote, arg_ref, arg_installation);
+
+  system = dir_get_system (arg_installation, &error);
+  if (system == NULL)
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (*arg_remote == 0 || strchr (arg_remote, '/') != NULL)
+    {
+      g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
+                                             "Invalid remote name: %s", arg_remote);
+      return TRUE;
+    }
+
+  if (!flatpak_dir_ensure_repo (system, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (!flatpak_dir_remove_ref (system, arg_remote, arg_ref, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  flatpak_system_helper_complete_remove_local_ref (object, invocation);
+
+  return TRUE;
+}
+
+static gboolean
+handle_prune_local_repo (FlatpakSystemHelper   *object,
+                         GDBusMethodInvocation *invocation,
+                         const gchar           *arg_installation)
+{
+  g_autoptr(FlatpakDir) system = NULL;
+  g_autoptr(GError) error = NULL;
+
+  g_debug ("PruneLocalRepo %s", arg_installation);
+
+  system = dir_get_system (arg_installation, &error);
+  if (system == NULL)
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (!flatpak_dir_ensure_repo (system, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  if (!flatpak_dir_prune (system, NULL, &error))
+    {
+      g_dbus_method_invocation_return_gerror (invocation, error);
+      return TRUE;
+    }
+
+  flatpak_system_helper_complete_prune_local_repo (object, invocation);
+
+  return TRUE;
+}
+
 static gboolean
 flatpak_authorize_method_handler (GDBusInterfaceSkeleton *interface,
                                   GDBusMethodInvocation  *invocation,
@@ -658,6 +1066,37 @@ flatpak_authorize_method_handler (GDBusInterfaceSkeleton *interface,
 
       action = "org.freedesktop.Flatpak.configure-remote";
 
+      polkit_details_insert (details, "remote", remote);
+    }
+  else if (g_strcmp0 (method_name, "Configure") == 0)
+    {
+      const char *key;
+
+      g_variant_get_child (parameters, 1, "&s", &key);
+
+      action = "org.freedesktop.Flatpak.configure";
+
+      polkit_details_insert (details, "key", key);
+    }
+  else if (g_strcmp0 (method_name, "UpdateRemote") == 0)
+    {
+      const char *remote;
+
+      g_variant_get_child (parameters, 1, "&s", &remote);
+
+      action = "org.freedesktop.Flatpak.update-remote";
+
+      polkit_details_insert (details, "remote", remote);
+    }
+  else if (g_strcmp0 (method_name, "RemoveLocalRef") == 0 ||
+           g_strcmp0 (method_name, "PruneLocalRepo") == 0)
+    {
+      const char *remote;
+
+      g_variant_get_child (parameters, 0, "&s", &remote);
+
+      action = "org.freedesktop.Flatpak.modify-repo";
+
       polkit_details_insert (details, "remote", remote);
     }
 
@@ -702,6 +1141,8 @@ on_bus_acquired (GDBusConnection *connection,
 
   helper = flatpak_system_helper_skeleton_new ();
 
+  flatpak_system_helper_set_version (FLATPAK_SYSTEM_HELPER (helper), 1);
+
   g_object_set_data_full (G_OBJECT(helper), "track-alive", GINT_TO_POINTER(42), skeleton_died_cb);
 
   g_dbus_interface_skeleton_set_flags (G_DBUS_INTERFACE_SKELETON (helper),
@@ -712,6 +1153,10 @@ on_bus_acquired (GDBusConnection *connection,
   g_signal_connect (helper, "handle-uninstall", G_CALLBACK (handle_uninstall), NULL);
   g_signal_connect (helper, "handle-install-bundle", G_CALLBACK (handle_install_bundle), NULL);
   g_signal_connect (helper, "handle-configure-remote", G_CALLBACK (handle_configure_remote), NULL);
+  g_signal_connect (helper, "handle-configure", G_CALLBACK (handle_configure), NULL);
+  g_signal_connect (helper, "handle-update-remote", G_CALLBACK (handle_update_remote), NULL);
+  g_signal_connect (helper, "handle-remove-local-ref", G_CALLBACK (handle_remove_local_ref), NULL);
+  g_signal_connect (helper, "handle-prune-local-repo", G_CALLBACK (handle_prune_local_repo), NULL);
 
   g_signal_connect (helper, "g-authorize-method",
                     G_CALLBACK (flatpak_authorize_method_handler),
@@ -722,7 +1167,7 @@ on_bus_acquired (GDBusConnection *connection,
                                          "/org/freedesktop/Flatpak/SystemHelper",
                                          &error))
     {
-      g_warning ("error: %s\n", error->message);
+      g_warning ("error: %s", error->message);
       g_error_free (error);
     }
 }
@@ -770,7 +1215,7 @@ message_handler (const gchar   *log_domain,
 {
   /* Make this look like normal console output */
   if (log_level & G_LOG_LEVEL_DEBUG)
-    g_printerr ("XA: %s\n", message);
+    g_printerr ("F: %s\n", message);
   else
     g_printerr ("%s: %s\n", g_get_prgname (), message);
 }
