@@ -4,17 +4,32 @@ set -e
 
 DIR=`mktemp -d`
 
+REPONAME=$1
+shift
 ID=$1
 shift
+COLLECTION_ID=$1
+shift
 
-flatpak build-init ${DIR} ${ID} ${ID} ${ID}
-sed -i s/Application/Runtime/ ${DIR}/metadata
+mkdir ${DIR}/files
+mkdir ${DIR}/usr
+cat > ${DIR}/metadata <<EOF
+[Runtime]
+name=${ID}
+EOF
+
+cat ${DIR}/metadata
 
 # Add bash and dependencies
 mkdir -p ${DIR}/usr/bin
 mkdir -p ${DIR}/usr/lib
 ln -s ../lib ${DIR}/usr/lib64
 ln -s ../lib ${DIR}/usr/lib32
+if test -f /sbin/ldconfig.real; then
+    cp /sbin/ldconfig.real ${DIR}/usr/bin/ldconfig
+else
+    cp `which ldconfig` ${DIR}/usr/bin
+fi
 T=`mktemp`
 for i in $@; do
     I=`which $i`
@@ -57,6 +72,12 @@ done
 mkdir -p ${DIR}/usr/lib/locale/
 cp -r /usr/lib/locale/C.* ${DIR}/usr/lib/locale/en_US
 
+if [ x$COLLECTION_ID != x ]; then
+    collection_args=--collection-id=${COLLECTION_ID}
+else
+    collection_args=
+fi
+
 mkdir -p repos
-flatpak build-export --runtime ${GPGARGS-} repos/test ${DIR}
+flatpak build-export ${collection_args} --runtime ${GPGARGS-} repos/${REPONAME} ${DIR}
 rm -rf ${DIR}
