@@ -22,7 +22,7 @@ set -euo pipefail
 . $(dirname $0)/libtest.sh
 
 skip_without_bwrap
-skip_without_user_xattrs
+[ x${USE_SYSTEMDIR-} != xyes ] || skip_without_user_xattrs
 
 echo "1..2"
 
@@ -32,9 +32,12 @@ make_extension () {
 
     local DIR=`mktemp -d`
 
-    ${FLATPAK} build-init ${DIR} ${ID} ${ID} ${ID}
-    sed -i s/Application/Runtime/ ${DIR}/metadata
+    cat > ${DIR}/metadata <<EOF
+[Runtime]
+name=${ID}
+EOF
     mkdir -p ${DIR}/usr
+    mkdir -p ${DIR}/files
     touch ${DIR}/usr/exists
     touch ${DIR}/usr/extension-$ID:$VERSION
 
@@ -87,8 +90,8 @@ EOF
 
 mkdir -p repos
 ostree init --repo=repos/test --mode=archive-z2
-. $(dirname $0)/make-test-runtime.sh org.test.Platform bash ls cat echo readlink > /dev/null
-. $(dirname $0)/make-test-app.sh > /dev/null
+. $(dirname $0)/make-test-runtime.sh test org.test.Platform "" bash ls cat echo readlink > /dev/null
+. $(dirname $0)/make-test-app.sh test "" > /dev/null
 
 # Modify platform metadata
 ostree checkout -U --repo=repos/test runtime/org.test.Platform/${ARCH}/master platform
